# -*- coding: utf-8 -*-
'''
    (C) 2012-2013 Mike Miller.
    License: GNU GPLv3+

    Set up and configure logging and custom levels.
'''
import sys, os, subprocess
import logging, logging.handlers
from datetime import datetime as dt
from logging import Formatter
from os.path import expanduser
from glob import glob
try:
    import curses
except ImportError:
    curses = None
    try:
        import colorama
        colorama.init()
    except ImportError:
        print '\nError: color support not available.  Install colorama:'
        print '       pip install colorama\n'


# level constants
logging.NOTE = 32       # positive yet important
logging.SKIP = 22       # skipped an action
logging.CHANGE = 27     # changed target
logging.EXCEPT = 42     # pave/lib exception

log = logging.getLogger()
out = sys.__stdout__
logfmt = '  %(levelname)-7.7s %(module)s: %(message)s'
logfmtf = '%(asctime)s' + logfmt[1:], '%H:%M:%S'
ext_font = glob(expanduser('~/.fonts/[Ss]ymbola*.???'))


# logging classes
class SecshFilterI(logging.Filter):
    'Filter annoying "Secsh channel..." info messages.'
    def filter(self, record):
        allow = True
        if (isinstance(record.msg, basestring) and
            record.msg.startswith('Secsh channel ')):
            allow = False
        if record.module == 'sftp':
            allow = False
        return allow


class SecshFilterD(logging.Filter):
    'Filter annoying "[chan X]..." debug messages.'
    def filter(self, record):
        allow = True
        if isinstance(record.msg, basestring):
            allow = not (record.module == 'channel')
        return allow


class EspeakHandler(logging.StreamHandler):
    'Speak messages for notification.'
    def __init__(self, *args, **kwargs):
        super(EspeakHandler, self).__init__(*args, **kwargs)
        with open(os.devnull, "w") as null:
            try:
                self.process = subprocess.Popen(['espeak'],
                    stdin=subprocess.PIPE, stderr=null )
            except:
                log.error('espeak is not installed.')
                self.process = None

    def emit(self, record):
        if self.process:
            try:
                msg = self.format(record)
                self.process.stdin.write(msg + '\n')
                self.process.stdin.flush()
            except:
                self.handleError(record)

    def __del__(self):
        if self.process:
            self.process.close()


class ColorFmtr(Formatter):
    'Colors the levelname of a log message.'
    def __init__(self, *args, **kwargs):
        #~ super(ColorFmtr, self) -->
        # Py 2.6: TypeError: super() argument 1 must be type, not classobj
        Formatter.__init__(self, *args, **kwargs)
        if curses:
            curses.setupterm()
            if curses.tigetnum('colors') < 8:
                raise EnvironmentError, 'Not enough colors available.'
            self.curses = curses
            setf = curses.tigetstr('setaf')
            setbg = curses.tigetstr('setab')
            bold = curses.tigetstr('bold')
            white_on_red = (curses.tparm(setf, curses.COLOR_WHITE) +
                            curses.tparm(bold, curses.A_BOLD) +
                            curses.tparm(setbg, curses.COLOR_RED))
            self.colormap = dict(
                DEBUG    = curses.tparm(setf, curses.COLOR_BLUE),
                INFO     = curses.tparm(setf, curses.COLOR_GREEN),
                SKIP     = curses.tparm(setf, curses.COLOR_CYAN),
                CHANGE   = curses.tparm(setf, curses.COLOR_YELLOW),
                WARNING  = curses.tparm(setf, curses.COLOR_YELLOW) +
                            curses.tparm(bold, curses.A_BOLD),
                NOTE     = curses.tparm(setf, curses.COLOR_GREEN) +
                            curses.tparm(bold, curses.A_BOLD),
                ERROR    = curses.tparm(setf, curses.COLOR_RED) +
                            curses.tparm(bold, curses.A_BOLD),
                EXCEPT   = curses.tparm(setf, curses.COLOR_RED) +
                            curses.tparm(bold, curses.A_BOLD),
                CRITICAL = white_on_red,
                FATAL    = white_on_red,
                NOTSET   = curses.tigetstr('sgr0')
            )
            self.icomap = dict(
                DEBUG       = u'•',
                INFO        = u'✓',
                SKIP        = u'⏬',
                CHANGE      = u'✎',
                WARNING     = u'⚠',
                NOTE        = u'★',
                ERROR       = u'✗',
                EXCEPT      = u'💣' if ext_font else u'✘',
                CRITICAL    = u'💀' if ext_font else u'✘',
                FATAL       = u'💀' if ext_font else u'✘',
                NOTSET      = u'•',
            )
        else:
            self.colormap = dict(
                DEBUG    = '\x1b[34m',                  # blue
                INFO     = '\x1b[32m',                  # green
                SKIP     = '\x1b[36m',                  # cyan
                CHANGE   = '\x1b[33m',                  # yellow
                WARNING  = '\x1b[33m\x1b[1m',           # yellow bold
                NOTE     = '\x1b[32m\x1b[1m',           # green bold
                ERROR    = '\x1b[31m\x1b[1m',           # red bold
                EXCEPT   = '\x1b[31m\x1b[1m',           # red bold
                CRITICAL = '\x1b[37m\x1b[1m\x1b[41m',   # white bold on red
                FATAL    = '\x1b[37m\x1b[1m\x1b[41m',   # white bold on red
                NOTSET   = '\x1b[0m' # '\x1b(B\x1b[m'   # reset
            )
            self.icomap = dict(
                DEBUG       = '-',
                INFO        = '+',
                SKIP        = 'v',
                CHANGE      = '~',
                WARNING     = '!',
                NOTE        = '@',
                ERROR       = 'x',
                EXCEPT      = 'X',
                CRITICAL    = '*',
                FATAL       = '*',
                NOTSET      = ' ',
            )

    def format(self, record):
        colormap = self.colormap
        levelname = record.levelname
        s = Formatter.format(self, record)
        ico = self.icomap[levelname]
        # have to replace afterward because "%(levelname)-7.7s" cuts esc seqs
        # UnicodeDecodeError here = unicode char in msg w/o leading u'...'
        s = s.replace(levelname, (colormap.get(levelname, '') +
                      ico + ' ' + levelname + colormap.get('NOTSET')), 1)
        return s


class MsecFormatter(Formatter):
    'Includes milliseconds in asctime.'
    converter=dt.fromtimestamp
    def formatTime(self, record, datefmt=None):
        ct = self.converter(record.created)
        t = ct.strftime(datefmt or "%Y-%m-%d %H:%M:%S")
        s = "%s.%03d" % (t, record.msecs)
        return s


class ioLogger:
    '''
        Redirect output streams to logging system.
            orig_stdout = sys.stdout                        # save for later
            sys.stdout = ioLogger(log, 'debug')             # redirect stdout
            # ...
            sys.stdout = orig_stdout                        # restore
    '''
    def __init__(self, logger, level='debug'):
        self.logger = logger
        self.level = logging._levelNames[level.upper()]
        self.buffer = u''

    def isatty(self):
        return False

    def flush(self, interactive=False):
        if interactive:
            self.logger.log(self.level, self.buffer)
            self.buffer = u''  # might help

    def write(self, text):
        # this could be refactored, but may be a bit faster this way.
        if text.endswith('\n'):     # flush buffer
            try:  text = self.buffer + text.rstrip()
            except UnicodeDecodeError:
                text = (self.buffer +
                        text.decode('utf8', errors='replace').rstrip())
            self.logger.log(self.level, text)
            self.buffer = u''
        else:                       # add to buffer
            try: self.buffer += text
            except UnicodeDecodeError:
                self.buffer += text.decode('utf8', errors='replace')
            # having a weird error with pyemphem, what are these chars?
            # looks like apostrophe and acute accent:
            # libastro-3.7.5/magdecl.c: In function ���E0000���:

# functions
def start_file_log(filepath):
    '''Log each host's actions to files in temp folder.'''
    h = logging.handlers.RotatingFileHandler(filepath,
                                             encoding='utf8', backupCount=5)
    h.doRollover()  # rollover each execution
    h.setFormatter(MsecFormatter(*logfmtf))
    h.setLevel(logging.DEBUG)
    h.addFilter(SecshFilterD())  # doesn't seem to be working :/
    log.addHandler(h)
    return h


def makefunc(level):
    def thefunc(msg, *args, **kwargs):
        if log.isEnabledFor(level):
            log._log(level, msg, args, **kwargs)
    return thefunc


# allows modules to use functions that log as if they were local :/
def _makeRecord(self, name, level, fn, lno, msg, args, exc_info, func=None,
                extra=None):
    '''Custom Logger.makeRecord that doesn't raise KeyError on extra.'''
    rv = logging.LogRecord(name, level, fn, lno, msg, args, exc_info, func)
    if extra is not None:
        for key in extra:
            rv.__dict__[key] = extra[key]
    return rv
logging.Logger.makeRecord = _makeRecord


# set up logging, custom levels
logging.addLevelName(logging.NOTE, 'NOTE')      # new levels
logging.addLevelName(logging.SKIP, 'SKIP')
logging.addLevelName(logging.CHANGE, 'CHANGE')
logging.addLevelName(logging.EXCEPT, 'EXCEPT')
logging.addLevelName(logging.CRITICAL, 'FATAL') # rename existing

# add convenience funcs
log.note    = makefunc(logging.NOTE)
log.skip    = makefunc(logging.SKIP)
log.change  = makefunc(logging.CHANGE)
log.exc     = makefunc(logging.EXCEPT)

console = logging.StreamHandler(sys.stdout)
log.addHandler(console)
log.setLevel(logging.DEBUG)
console.addFilter(SecshFilterI())




'''
    (C) 2012-2013 Mike Miller.
    License: GNU GPLv3+

    Strangely enough, schema and schema-related code lives here.
'''
import imp, logging, unicodedata
from os.path import dirname, join

import voluptuous
from voluptuous import All, Any, Required as req, Range as rng, Invalid

import lib
from utils import shplit

log = logging.getLogger()
CURRENT = 'CURRENT'

# classes
class SchemaError(Exception):
    pass

def if_var_three_args(v):
    if not isinstance(v, basestring):
        raise Invalid('not a string.')
    args = shplit(v)
    if len(args) != 3:
        raise Invalid('variable test: wrong number of args: %r' % v)


# module schema additions
conditionals_schema = {
    'if':                               if_var_three_args,
    'if-exec':                          basestring,
    'if-platform':                      basestring,
    'do':                               basestring,
}

# define core section schemas
main_schema = {
    req('bak-files', default=True):     bool,
    req('env', default={}):             dict,
    req('include', default=[]):         Any(basestring, [basestring]),
    req('inspect', default=True):       Any(bool, basestring),
    req('jobs', default=1):             All(int, rng(min=1,max=100)),
    req('log-to', default=''):          basestring,
    req('sudo', default=True):          bool,
    'sudo-user':                        basestring,
    'sys.path':                         Any(basestring, [basestring]),
    req('targets', default=''):         Any(basestring, [basestring]),
    req('user', default=CURRENT):       basestring,
    req('version', default=CURRENT):    basestring,
    req('warn-only', default=False):    bool,
}

def validate_varname(value, msg=None):
    'Rules for variable names in the pavefile.'
    msg = 'variable names start with letters, or underscore.'
    cat = unicodedata.category(value[0])
    if not cat in ('Ll', 'Lu', 'Pc'):
        raise Invalid(msg)
    if cat == 'Pc' and value[0] != '_':
        raise Invalid(msg)

    msg = 'variable names must be with letters, digits, or underscore.'
    for ch in value[1:]:
        cat = unicodedata.category(ch)
        if not cat in ('Ll', 'Lu', 'Pc', 'Nd'):
            raise Invalid(msg)
        if cat == 'Pc' and ch != '_':
            log.note('raising...')
            raise Invalid(msg)
    return value


vars_schema = {
    validate_varname:   Any(int, float, basestring, bool, list),
}
target_groups_schema = {
    basestring:         Any(basestring, [basestring]),
}
passwords_schema = [basestring]

# load library module schemas
mod_schemas = {}
for modname in [ m for m in lib.modlist ]:
    modpath = join(dirname(__file__), 'lib', '%s.py' % modname)
    mod = imp.load_source('pave.lib.' + modname, modpath)
    if mod and hasattr(mod, 'schema'):
        mod_schemas[modname] = mod.schema

tasks_schema = [basestring, mod_schemas]
fabtasks_schema = [basestring, dict]
disabled_schema = fabtasks_schema

root_schema = {
    req('main'):        main_schema,
    'passwords':        passwords_schema,
    'target-groups':    target_groups_schema,
    'tasks':            tasks_schema,
    'disabled':         disabled_schema,
    'fab-tasks':        fabtasks_schema,
    'vars':             vars_schema,
}
schema = voluptuous.Schema(root_schema, extra=False)


if __name__ == '__main__':

    from pprint import pprint
    pprint(schema)




# -*- coding: utf-8 -*-
# (C) 2012-2013 Mike Miller.
# License: GNU GPLv3+
'''
    Manage software packages.
    Packages names for ``install``, ``remove``, etc. are listed as
    whitespace-delimited strings.

    ``expiration:``
        The age in days after which the OS package index is considered
        to have expired, requiring action.

        To be a good neighbor, the default, 0 (= 24 hours) means packages
        won't be updated unless it has been a full day since the last update.
        A value of -1 forces action.
    ``inventory:``
        Make a list of installed packages to compare against.
    ``update/upgrade:``
        Update the system's package index, upgrade packages to latest release.
    ``upgrade-full:``
        Upgrade packages, even if it causes obsolete packages to be removed,
        or new packages to be installed, for example new kernel versions.
        e.g. Debian's "dist-upgrade".

    Example::

        - packages:
            if-platform: Debian
            install:
                sysvbanner python-pip
            # expiration: -1    # force attempt

'''
import logging

from fabric.api import env
from fabric.context_managers import hide
from voluptuous import Required as req

from pave.schema import conditionals_schema
from pave.utils import eval_conditionals, err_cond, runcmd

log = logging.getLogger()
schema = {
    req('expiration',   default=0):         int,
    req('install',      default=''):        basestring,
    req('inventory',    default=True):      bool,
    req('remove',       default=''):        basestring,
    req('update',       default=True):      bool,
    req('upgrade',      default=True):      bool,
    req('upgrade-full', default=True):      bool,
}
schema.update(conditionals_schema)


def cleanup(cmds, context):
    'Clean any mess left over by package installation.'
    changed, errors, removed = 0, 0, 0
    log.debug(u'package.cleanup…')
    result = runcmd(cmds.pkg_clean, **context)
    if result.failed:
        log.error(result); errors = 1
        if env.pave_raise_errs: raise RuntimeError

    elif hasattr(cmds, 'parse_pkg_output'):
        modifications = cmds.parse_pkg_output(result)
        removed = modifications[2]
        if any(modifications):
            changed += 1
    return changed, errors, removed


def _check_err(result):
    errs = 0
    if result.failed:
        log.error(result); errs += 1
        if env.pave_raise_errs: raise RuntimeError
    return errs


def handle(section, cmds, context):
    'Run the system package manager, and any others given.'
    changed, errors = 0, 0
    parse = hasattr(cmds, 'parse_pkg_output')
    if eval_conditionals(section, context) is False:
        log.skip(err_cond, '')  # log here to tag from module
        return changed, errors

    # inventory existing packages
    inventory, pkginv = section['inventory'], [] # schema default
    log.debug('inventory: %s', inventory)
    if inventory:
        log.info('inventory: collecting installed pkg list.')
        with hide('stdout'):
            pkginv = runcmd(cmds.pkg_inventory)
            if pkginv.failed:
                log.error(pkginv); errors += 1
                if env.pave_raise_errs: raise RuntimeError
            pkginv = pkginv.split()
        if log.isEnabledFor(logging.DEBUG):
            log.debug('found %s packages.', len(pkginv))

    # update index
    update = section['update']
    expires = section['expiration']
    log.debug('update: %s  expiration: %s day(s)', update, expires)
    if update:
        # skip if done recently
        result = runcmd(cmds.pkg_update_time, expires, **context)
        if result.failed:
            log.error(result); errors += 1 # don't halt on test failure
        if result:  # string returned == is old enough
            log.change(u'update: OS pkg index… (takes a few mins).')
            result = runcmd(cmds.pkg_update, **context)
            if result.succeeded:   changed += 1
            else:
                log.error(result); errors += 1
                if env.pave_raise_errs: raise RuntimeError
        else:
            log.skip(u'update: occurred recently.')

    # save last pkg modification time
    if not parse:
        lasttime = runcmd(cmds.pkg_stat_mtime, **context)
        errors += _check_err(lasttime)

    def handle_operation(cmd, name, text):
        ''' Repetitive code for each package-change operation.
            There are currently two strategies on how to discern if an
            upgrade happened.  Reading the output of the command, or
            checking mtime on the filesystem.  Not clear which is portable.
        '''
        chd, errs = 0, 0
        # try the operation
        result = runcmd(cmd, **context)
        errs += _check_err(result)

        if parse:
            changes = cmds.parse_pkg_output(result)
            log.debug('parse: output %s', changes)
            changes = any(changes)
        else:
            thistime = runcmd(cmds.pkg_stat_mtime, **context)  # look again
            errs += _check_err(thistime)
            if str(lasttime) == str(thistime):
                changes = True

        if changes:
            log.change('%s: %s.', name, text);  chd += 1
        else:
            log.skip('%s: %snot %s.', name, (inventory or 'oops '), text)
        return chd, errs

    # remove: should be before upgrade, so removed aren't upgraded.
    pkgstr = section['remove']
    log.debug('remove: %r', pkgstr)
    if pkgstr:
        if inventory:
            pkgstr = ' '.join([ p  for p in pkgstr.split()  if p in pkginv ])
        if pkgstr:
            log.info(u'remove: %s…', pkgstr)
            cmdstr = cmds.pkg_remove % pkgstr
            chd, errs = handle_operation(cmdstr, 'remove', str(pkgstr))
            changed += chd; errors += errs
        else:
            log.skip(u'remove: NADA—inventory complete.')

    # upgrade
    upgrade = section['upgrade']
    log.debug('upgrade: %s', upgrade)
    if upgrade:
        # skip if done recently
        result = runcmd(cmds.pkg_upgrd_time, expires, **context)
        if result.failed:
            log.error(result); errors += 1 # don't halt on test failure

        if result:  # True/string == is old enough
            log.info(u'upgrade: OS pkgs… (may take a few mins).')
            if (section.get('upgrade-full') and
                getattr(cmds, 'pkg_upgrd_full', None)):
                upgrd_cmd = cmds.pkg_upgrd_full
            else:
                upgrd_cmd = cmds.pkg_upgrade

            chd, errs = handle_operation(upgrd_cmd, 'upgrade', 'done')
            changed += chd; errors += errs
        else:
            log.skip(u'upgrade: occurred recently.')

    # install
    pkgstr = section['install']
    log.debug('install: %r', pkgstr)
    if pkgstr:
        if inventory:
            newpkgs = [ p for p in pkgstr.split() if p not in pkginv ]
            pkgstr = ' '.join(newpkgs)
        if pkgstr:
            log.info(u'install: %s pkgs… (may take a few mins).',
                      len(pkgstr.split()))
            cmdstr = cmds.pkg_install % pkgstr
            chd, errs = handle_operation(cmdstr, 'install', str(pkgstr))
            changed += chd; errors += errs
        else:
            log.skip(u'install: NADA—inventory complete.')

    return int(bool(changed)), errors  # only one change




# -*- coding: utf-8 -*-
# (C) 2012-2013 Mike Miller.
# License: GNU GPLv3+
'''
    Executes a list of commands using the remote shell.  Each command may be:

    * A string
        A command that runs under the default user.
    * A mapping with one or more of the following members:
        | ``do: CMD_STRING``
        |     The command to run.
        | ``user: USERNAME``
        |     Run as another user.
        |     May be ``sudo`` (root), or a valid remote username.
        | ``title: TITLE``
        |     to document or hide long, ugly commands.
        | ``workdir: PATH`` set working directory.
        | ``local: <bool>`` runs it locally instead of remotely.

    See :ref:`cond_exec` for more details on ``if:`` and ``if-exec:``.

    .. note::

        While *not recommended* to set passwords via command-line tools for
        various reasons, tasks makes an attempt to hide them (vars starting
        with "``pwd_``") from standard logs.  Passwords will still be displayed
        in debug logs (and perhaps shell history) upon command-line execution.

    Example::

        - banner "Hello!"     # world

        # signal a change event and prevent errors from halting execution:
        - touch %filename && echo CHANGED || true
        - rm /etc/nginx/sites-enabled/default && echo CHANGED || echo SKIPPED

        # Even poor-er man's conditionals:
        - " [ '%mode' == 'qa' ] && echo 'QA install' "

        # more robust task definition:
        - run:
            - title: Nothing to see here...
              if-exec: test -f /foo
              do: echo "This is running under $USER (appuser)"
              user: appuser

'''
import logging

from fabric.api import env
from fabric.context_managers import cd
from voluptuous import Any

from pave.schema import conditionals_schema
from pave.utils import (runcmd, trunc, shell_error_map, sequence,
                        check_list_conditionals)

log = logging.getLogger()
subtask_schema = {
    'local': bool,
    'do': basestring,
    'user': basestring,
    'title': basestring,
    'workdir': basestring,
}
subtask_schema.update(conditionals_schema)

schema = [ Any(basestring, subtask_schema) ]


def make_title(title, command, pwds):
    if not title:
        title = trunc(command, 72).rstrip()
        for pwd in pwds:        # obscure passwords in std logs, broken?
            if pwd in title:
                title = title.replace(pwd, '****')
    return title


def handle(section, cmds, context, chatty=True):
    changed, errors = 0, 0
    context = context.copy()    # needs to be here; may be used by other mods
    context.chunking = False    # disable bracket splitting
    pwds = env.pave_passwords.keys()

    for i, command in enumerate(sequence(section)):
        title = ''
        local = None
        workdir = '.'
        mycon = context    # may change on each command and need to copy

        if type(command) is dict:
            title = command.get('title', title)
            workdir = command.get('workdir', workdir)
            local = command.get('local', None)
            user = command.get('user', None)

            cmdstr = command.get('do')
            if not cmdstr:
                log.error('task %s in section %s missing do: clause.',
                          i, __name__.split('.')[-1])
                errors += 1
                if env.pave_raise_errs:
                    raise RuntimeError

            mycon = context.copy()  # prevent context spill
            if local:
                mycon.local = True
            elif user == 'sudo':
                mycon.use_sudo = True
            elif user:
                mycon.use_sudo = True
                mycon.sudo_user = user

            # down here--tests need to be run as correct user, dir
            with cd(workdir):
                command = check_list_conditionals(command, mycon,
                                                module=__name__, replace=False)
            if not command: continue

            title = make_title(title, cmdstr, pwds)
        else:
            cmdstr = command
            title = make_title(title, cmdstr.partition('&&')[0], pwds)

        # execute
        with cd(workdir):
            if chatty:
                log.info(title if title.endswith(u'…') else title + u'…')
            result = runcmd(cmdstr, **mycon)
            if result.succeeded:
                if 'CHANGED' in result:             # if both, CHANGED wins
                    changed += 1
                    if chatty: log.change(title + u' » DONE.')
                elif result.endswith('SKIPPED'):
                    if chatty: log.skip(title)
                if local:
                    log.debug('local-output:\n' + result)
            else:
                msg = title + '\n            Full Error:\n'
                text = result.decode(context.encoding)
                log.debug(text)

                msg = title + '\n            Returned: '
                msg += '\n'.join( text.rstrip().rsplit('\n', 3)[1:] )
                msg += '\n            Exit code: %s, ' % result.return_code
                msg += shell_error_map.get(result.return_code, 'Unknown')
                log.error(msg); errors += 1
                if env.pave_raise_errs:
                    raise RuntimeError

    return changed, errors




# (C) 2012-2013 Mike Miller.
# License: GNU GPLv3+
'''
    Create operating system users.

    Example::

        - users:
            - name: appuser                 # required
              groups: appgroup www-data     # whitespace separated groups
              create-home: True
              shell: /bin/bash
              password: use appuser         # see passwords: section

'''
import logging

from fabric.api import env

from voluptuous import Any, Required as req
from pave.schema import conditionals_schema
from pave.utils import check_list_conditionals, find_password, runcmd

log = logging.getLogger()
user_schema = {
    req('name'):                            basestring,
    req('groups',       default=None):      Any(None, basestring),
    req('create-home',  default=True):      bool,
    'shell':                                basestring,
    'password':                             basestring,
}
user_schema.update(conditionals_schema)
schema = [user_schema]


def handle(section, cmds, context):
    changed, errors = 0, 0
    context.chunking = True  # needed by these commands

    existing = runcmd(cmds.usr_lst)  # doesn't need to run as root
    if existing.failed:  errors += 1
    if existing:
        existing = existing.split()
    log.debug('existing: %s', existing)

    for user in section:
        user = check_list_conditionals(user, context, module=__name__, replace=False,
                             force=True)
        if not user: continue

        mycon = context.copy()    # prevent context spills
        name = user.pop('name')
        if name in existing:
            log.skip('not creating existing user: %s', name)
        else:
            log.change('creating user: %s', name)
            # check defaults
            if user['groups']:  # comma delimited
                user['groups'] = ','.join(user['groups'].split())

            if 'password' in user:
                user['password'], myerr = find_password(user['password'], True)
                errors += myerr

            mycon.update(user)
            result = runcmd(cmds.usr_add, name, **mycon)

            if result.succeeded:   changed += 1
            else:
                log.error(result); errors += 1
                if env.pave_raise_errs: raise RuntimeError

    return changed, errors




# -*- coding: utf-8 -*-
# (C) 2012-2013 Mike Miller.
# License: GNU GPLv3+
'''
    Configure the remote system.  Also handles templating, see below.

    ``render:``
        Renders a template to a local file, which is then uploaded to host.

        ``engine:``
            The templating language, such as python (printf|Template|format),
            or the executable name of an external package, e.g: ``jinja2``

    ``render-remote:``
        Simple remote "templating" with sed.

    Example::

        - configure:            # str || list accepted on all sub-tasks below:
            create:  # file or folder/
                - foo.conf
                - /tmp/project/

            append:  # text remotefname
                Answer=42 %rc

            comment: # [comment char] from-regex remote-fname
                - ' "insecure on" %rcfile '

            replace: # from-regex to remote-fname
                - ^foo$ bar %rcfile
                # note that because of yaml syntax and splitting, two layers
                # of quoting or blocks may be needed when backslashes used:
                - >
                    '^foo\s*$' bar %rcfile

            render: # engine [extra-vars] local-template remote-fname
                - printf swappiness=50 examples/printf_template.txt %rc2
                - format FOO=BAR       examples/format_template.txt %rc3

            render-remote: # [vars] remote-source remote-dest
                - INSTANCE_NUM=1 %sitecfg /etc/init/%svcname.conf

            update: # copy a file into place if newer, SRC DEST
                - ~/%sitename/cfg/celeryd.conf /etc/init/celeryd.conf

            access: # mode user group remote-fname # use "" for empty
                - 0640 %user %user %rc2
'''
import os, logging, tempfile, traceback
from string import Template

from fabric.api import env
from fabric.operations import put
from fabric.contrib.files import sed, comment, exists as rexists
from voluptuous import Any, Required
try:
    import jinja2
except ImportError:
    jinja2 = None

from pave.utils import (runcmd, sequence, contains, backup, shplit, q,
                        eval_conditionals, err_cond)
from pave.schema import conditionals_schema


log = logging.getLogger()
fl = 20 # filename length
vsch = Any(basestring, [basestring])
req = lambda key: Required(key, default=[])
schema = {
    req('append'): vsch, req('comment'): vsch, req('replace'): vsch,
    req('render'): vsch, req('update'):  vsch, req('access'):  vsch,
    req('render-remote'): vsch, req('create'): vsch,
}
schema.update(conditionals_schema)


def _format_exc(text):
    return '\n          ' + '\n          '.join(text.split('\n'))


def render(engine, text, tvars):
    'Render templates.'
    from pave.main import tempdir # avoid circ. deps
    try:
        fd, tempfname = tempfile.mkstemp(dir=tempdir)
        log.debug('created tempfile at %s', tempfname)

        if engine == 'printf':
            result = text % tvars
        elif engine == 'format':
            result = text.format(**tvars)
        elif engine == 'Template':
            result = Template(text).substitute(**tvars)
        elif engine == 'jinja':
            if jinja2:  result = jinja2.Template(text).render(**tvars)
            else:       log.error('Jinja templating engine not installed!')

        if log.isEnabledFor(logging.DEBUG):
            log.debug(repr(result))
        os.write(fd, result.encode('utf8'))
        os.close(fd)
        return tempfname
    except KeyError, e:
        log.error('Missing given variable: %r', e.message)
    except Exception, e:
        log.error(_format_exc(traceback.format_exc(limit=5)))


def handle(section, cmds, context):
    # needs factoring into sep functions
    changed, errors = 0, 0
    if eval_conditionals(section, context) is False:
        log.skip(err_cond, '')  # log here to tag from this module
        return changed, errors

    context = context.copy()  # needs to be here; may be used by other mods
    context.pop('sudo_user', None)
    context.pop('encoding', None)
    context.pop('chunking', None)

    sections = sorted(section.keys())
    if 'access' in sections:            # move access to the back
        sections.remove('access')
        sections.append('access')

    for operation in sections:
        for task in sequence(section[operation]):
            log.debug(u'splitting %r…', task)     # to pinpoint shlex errs
            args = shplit(task)

            if operation == 'append':
                text, fname = args
                if contains(fname, text, **context):
                    log.skip(u'append: text %.20r… exists in "%s".', text,
                              fname[-fl:])
                else:
                    if rexists(fname):
                        # need to reimplement contains to avoid 2x exist check
                        if env.pave_bak: backup(fname, cmds, context)
                    log.change(u'appending text %.20r… to "…%s".', text,
                                fname[-fl:])
                    # more functional/efficient to implement append ourselves
                    if "'" in text: text = text.replace("'", r"\'")
                    result = runcmd(cmds.file_append, text, fname, **context)
                    if result.succeeded:   changed += 1
                    else:
                        log.error(result); errors += 1
                        if env.pave_raise_errs: raise RuntimeError

            elif operation == 'create':
                fname = args[0]
                if rexists(fname):
                    log.skip(u'create: %.40r… exists.', fname)
                else:
                    log.change(u'creating %.40r…', fname)
                    obj_create = (cmds.dir_create  if fname.endswith('/')
                                                   else cmds.file_create)
                    result = runcmd(obj_create, fname, **context)
                    if result.succeeded:   changed += 1
                    else:
                        log.error(result); errors += 1
                        if env.pave_raise_errs: raise RuntimeError

            elif operation == 'comment':  # comment and replace roughly the same
                if len(args) == 3:  char = args[-3]
                else:               char = '# '
                text, fname = args[-2], args[-1]
                after = char + text

                if contains(fname, after, **context):
                    log.skip(u'comment: "%s" has comment "%.10s…".',
                              fname[-fl:], after)
                else:
                    if contains(fname, text, **context):
                        if env.pave_bak: backup(fname, cmds, context)
                        log.change(u'commenting %.20r… in "%s".', text,
                                    fname[-fl:])
                        result = comment(fname, text, char=char, **context)
                        if result.succeeded:   changed += 1
                        else:
                            log.error(result); errors += 1
                            if env.pave_raise_errs: raise RuntimeError
                    else:
                        log.error(u'comment: text %.20r… not found in "%s".',
                                  text, fname[-fl:])
                        errors += 1
                        if env.pave_raise_errs: raise RuntimeError

            elif operation == 'replace':
                # this operation isn't very efficient
                before, after, fname = args
                if (contains(fname, after, fixed=True, **context) and
                    # when after is a subset of before
                    not contains(fname, before, **context)):
                    log.skip(u'replace: "%s" has text %.20r… .', fname[-fl:],
                              after)
                else:
                    if contains(fname, before,  **context):
                        if env.pave_bak: backup(fname, cmds, context)
                        log.change(u'replacing %.20r… in "%s".', before,
                                    fname[-fl:])
                        result = sed(fname, before, after,
                            backup=('' if env.pave_bak else '.bak'), **context)
                        if result.succeeded:   changed += 1
                        else:
                            log.error(result); errors += 1
                            if env.pave_raise_errs: raise RuntimeError
                    else:
                        log.error(u'replace: %.20r… not found in "%s".',
                                   before, fname[-fl:])
                        errors += 1
                        if env.pave_raise_errs: raise RuntimeError

            elif operation == 'render':  # check local file with schema
                remotefname = args[-1]
                if rexists(remotefname, **context):
                    log.skip('render: "%s" already exists.', remotefname[-fl:])
                else:
                    # no backup needed here
                    log.change('rendering to "%s".', remotefname[-fl:])
                    eng = args[0]
                    tvars = [ arg.split('=',1) for arg in args[1:-2] if '=' in arg ]
                    tmplfname = args[-2]

                    # local should overwrite global vars
                    tvars = dict(tvars)
                    copy = env.pave_vars.copy()
                    copy.update(tvars)
                    tvars = copy

                    with file(tmplfname) as f:
                        tempfname = render(eng, f.read(), tvars)
                        result = put(tempfname, remotefname)
                        if result.succeeded:
                            changed += 1
                            os.unlink(tempfname) # rm
                        else:
                            log.error(result); errors += 1
                            if env.pave_raise_errs: raise RuntimeError

            elif operation == 'render-remote':
                src, dest = args[-2], args[-1]

                result = runcmd(cmds.test_nt, src, dest, **context)
                if result.succeeded:
                    log.change(u'render: copying newer "…%s" to "…%s".',
                                src[-fl:], dest[-fl:])
                    command = 'sed '
                    tvars = [ arg.split('=',1) for arg in args[0:-2] if '=' in arg ]
                    for tvar in tvars:
                        command += "-e 's/%s/%s/g' " % tuple(tvar)
                    command += '%s > %s'

                    result = runcmd(command, q(src), q(dest), **context)
                    if result.succeeded:   changed += 1
                    else:
                        log.error(result); errors += 1
                        if env.pave_raise_errs: raise RuntimeError
                else:
                    log.skip(u'rend-rem: template "…%s" older than "…%s".',
                                src[-fl:], dest[-fl:])

            elif operation == 'update':
                src, dest = args

                result = runcmd(cmds.test_nt, src, dest, **context)
                if result.succeeded:
                    log.change(u'update: copying newer "…%s" to "…%s".',
                                src[-fl:], dest[:fl])

                    result = runcmd(cmds.file_update, q(src), q(dest),
                                    **context)
                    if result.succeeded:   changed += 1
                    else:
                        log.error(result); errors += 1
                        if env.pave_raise_errs: raise RuntimeError
                else:
                    log.skip(u'update: "…%s" older than "…%s".',
                                src[-fl:], dest[-fl:])

            elif operation == 'access':
                params = args[:-1]
                fn = args[-1]
                # get details
                results = runcmd(cmds.file_stat, fn)
                if results.succeeded:
                    cmode, cuser, cgrp = results.split()
                else:
                    log.error('access: unable to stat "%s"', fn[-fl:])
                    cmode, cuser, cgrp = '', '', ''
                    errors += 1; continue
                    # don't quit on test failure
                # no backups for metadata changes
                for i, param in enumerate(params):
                    if i == 0: # mode
                        mode = param.zfill(4)
                        cmode = cmode.zfill(4)
                        if mode == cmode:
                            log.skip('access: "%s" has mode %s', fn[-fl:], mode)
                        else:
                            log.change('access: changing "%s" to mode %s', fn[-fl:], mode)
                            res = runcmd(cmds.file_chmod, mode, fn, **context)
                            if res.succeeded:   changed += 1
                            else:
                                log.error(res); errors += 1
                                if env.pave_raise_errs: raise RuntimeError

                    elif i == 1: # user
                        if param == cuser:
                            log.skip('access: "%s" has owner %s', fn[-fl:], param)
                        else:
                            log.change('access: changing "%s" to owner %s', fn[-fl:], param)
                            res = runcmd(cmds.file_chown, param, fn, **context)
                            if res.succeeded:   changed += 1
                            else:
                                log.error(res); errors += 1
                                if env.pave_raise_errs: raise RuntimeError

                    elif i == 2: # group
                        if param == cgrp:
                            log.skip('access: "%s" has group %s', fn[-fl:], param)
                        else:
                            log.change('access: changing "%s" to group %s', fn[-fl:], param)
                            res = runcmd(cmds.file_chgrp, param, fn, **context)
                            if res.succeeded:   changed += 1
                            else:
                                log.error(res); errors += 1
                                if env.pave_raise_errs: raise RuntimeError
            else:
                log.warn('unknown operation "%s" given.')

    return changed, errors




# -*- coding: utf-8 -*-
# (C) 2012-2013 Mike Miller.
# License: GNU GPLv3+
r'''
    Deploy code with this version control helper.
    $VCS support must be installed beforehand, see the `packages`_ section.

    If the destination does not exist, an initial checkout is done.
    If it does exist, the repository is updated to the latest version.

    Formats::

        - VCS_TYPE REPO_URL DEST EXTRA          # simple or
        - do: VCS_TYPE REPO_URL DEST EXTRA      # robust

    ``VCS_TYPE``:
        Version Control System, one of ``svn``, ``hg``, or ``git``.
    ``REPO_URL``:
        URL to repository.
    ``DEST``:
        Destination folder.
    ``EXTRA``:
        Extra command line parameters.
    ``do``:
        What to do.
    ``input``:
        Useful to automate interactive prompts.
    ``user:``:
        Execute vcs under this remote username.

    Example::

        - vcs:
            - do: hg %repo_base/appname ~%user/appname
              input: p\nn                           # interactive answer
              user: %user

'''
from os.path import dirname, basename, join
import logging

from fabric.contrib.files import exists as rexists
from voluptuous import Invalid, Any

from pave.schema import conditionals_schema
from pave.utils import shplit, check_list_conditionals


log = logging.getLogger()
pwds = []
vcmap = {
    'svn':  dict(checkout='checkout', update='update'),
    'hg':   dict(checkout='clone',    update='pull -u', chkupdate='incoming'),
    'git':  dict(checkout='clone',    update='pull'),
}
vcslist = vcmap.keys()

# trying to figure out the best way of doing this:
def min_two_tokens(v):
    'validate'
    tokens = v.split()
    length = len(tokens)
    if length < 3:
        raise  Invalid('Expected length of at least 3.')
    return v

def in_vclist(v):
    'validate'
    tokens = v.split()

    if tokens[0] not in vcslist:
        raise  Invalid('Unknown vcs: ', tokens[0])
    return v

#~ chkstr = All(basestring, min_two_tokens, in_vclist)  # broken :/
chkstr = basestring  # broken because of new input key
vcs_schema = {
    'input': basestring,
    'user': basestring,
}
vcs_schema.update(conditionals_schema)
schema = [Any(basestring, vcs_schema)]


def handle(section, cmds, context):
    changed, errors = 0, 0
    from pave.lib.run import handle as runhandle

    for i, directive in enumerate(section):
        inpstr = ''
        user = None
        mycon = context.copy()
        if type(directive) is dict:
            inpstr = directive.get('input', inpstr)
            user = directive.get('user', user)
            if not 'do' in directive:
                raise ValueError, 'no do: key.'  # should be caught in schema
            if user == 'sudo':
                mycon.use_sudo = True
            elif user:
                mycon.use_sudo = True
                mycon.sudo_user = user

        directive = check_list_conditionals(directive, mycon, module=__name__,
                                    replace=False)
        if not directive: continue

        tokens = shplit(directive.get('do'))
        dest = ('~' + user) if user else ''
        vcs = tokens[0]
        repo = tokens[1]
        reponame = (basename(repo) or 'repo') + ' ' # use repo if name lost
        if len(tokens) > 2:
            dest = join(dest, tokens[2])
        thisvcs = vcmap[vcs]
        exists = rexists(dest)

        if exists:
            log.info(u'%sexists, checking for updates…', reponame)
            if vcs == 'hg': # hg gives notice of updates on incoming, not pull
                command = '%s %s && %s %s' % (vcs, thisvcs['chkupdate'],
                                              vcs, thisvcs['update'])
            else:
                command = '%s %s' % (vcs, thisvcs['update'])
        else:               # checkout
            log.change(u'%smissing, performing checkout…', reponame)
            command = '%s %s %s %s' % (vcs, thisvcs['checkout'], repo, dest)
            dest = dirname(dest)

        if len(tokens) > 3:  # add any extra params to the end
            command += ' ' + ' '.join(tokens[3:])

        # need to feed input?
        if inpstr:
            command = ("echo -e '%s' | " % inpstr) + command

        # run each task separately so we can log the results as they happen.
        command += ' && echo CHANGED || true'
        task = {'do':command, 'workdir':dest, 'user':user}

        chd, errs = runhandle([task], cmds, mycon, chatty=False)
        if exists:
            if chd: log.change('repo updated.')
            else:   log.skip('no updates found.')
        changed += chd
        errors += errs

    return changed, errors




# -*- coding: utf-8 -*-
# (C) 2012-2013 Mike Miller.
# License: GNU GPLv3+
'''
    Manage postgres.
    *Captain Obvious here to chime-in that postgres must be installed on
    remote target beforehand, see* `packages`_.

    ``link current:``
        Create a symbolic link from the last sorted item in
        ``/etc/postgresql/`` to ``/etc/postgresql/curr``.
    ``replace``,  ``access:``
        See the corresponding directive under configure.
        These are for convenience to group postgres-related configuration
        together.
    ``user``,  ``db``,  ``grant``,  ``sql:``
        ``postgresql-client-common`` (or distro equiv.) must be installed on
        remote target beforehand.
        Run the appropriate commands to create objects or configure the
        database.  String or list of strings may be given.

    Example::

        - postgres:
            link current: False
            replace:
                - '#\s*password_enc password_enc %pgcfg/postgresql.conf'

            su-password:
                use pgsu                                # see passwords:
                # md5 acbd18db4cc2f85cedef654fccc4a4d8  # how to hard-code

            db: # quotes not required below, but fixes highlighting. :/
                - name: "%sitename"
            user:
                - name: "%sitename"
            grant:
                # privs   on_obj    named         to_user
                all       database  %sitename     %sitename

            sql: # str|list lines of sql, first token should be dbname:
                template1 select datname from pg_database
                          where datname = '%sitename';

'''
import logging

from fabric.api import env, run
from fabric.contrib.files import exists as rexists #append,
from voluptuous import Any, Required as req

from pave.lib.configure import handle as cfghandle, vsch as cfgvsch
from pave.schema import conditionals_schema
from pave.utils import (runcmd, sequence, eval_conditionals, err_cond,
                        find_password, md5, shplit)


log = logging.getLogger()

# section schema
createuser_schema = Any(None, [{
    req('name'):                          basestring,
    req('encrypted',    default=True):    bool,
    req('createdb',     default=False):   bool,
    req('createrole',   default=False):   bool,
    req('superuser',    default=False):   bool,
    req('login',        default=True):    bool,
}] )
createdb_schema = Any(None, [{
    req('name'):                          basestring,
    req('template', default='template0'): basestring,
    'tablespace':                         basestring,
    'encoding':                           basestring,
    'locale':                             basestring,
    'description':                        basestring,
}] )
schema = {
    req('link current', default=True):    bool,
    req('replace',      default=[]):      cfgvsch,
    req('access',       default=[]):      cfgvsch,
    req('su-password',  default=''):      basestring,  # need validation here
    req('user',         default=[]):      createuser_schema,
    req('db',           default=[]):      createdb_schema,
    req('grant',        default=[]):      Any(basestring, [basestring]),
    req('sql',          default=[]):      Any(basestring, [basestring]),
}
schema.update(conditionals_schema)


def handle(section, cmds, context):
    changed, errors = 0, 0
    if eval_conditionals(section, context) is False:
        log.skip(err_cond, '')  # log here to tag from module
        return changed, errors

    if section.pop('link current', None):
        # get pg versions
        result = run('ls -1 /etc/postgresql/')
        if result.succeeded:
            dirs = sorted(result.split('\n'))
            if dirs:
                latest = dirs[-1]
                if rexists('/etc/postgresql/curr'):
                    log.skip('/etc/postgresql/curr exists')
                else:
                    log.change('linking %s to ./curr', latest)
                    result = runcmd(cmds.pg_ln_curr, latest, **context)
                    if result.succeeded:   changed += 1
                    else:
                        log.error(result); errors += 1
                        if env.pave_raise_errs: raise RuntimeError
            else:
                log.error('link: config not found--is pg installed?'); errors += 1
                if env.pave_raise_errs: raise RuntimeError
        else:
            log.error('link: %s--is pg installed?', result); errors += 1
            if env.pave_raise_errs: raise RuntimeError

    # set this after link_curr, since it has to be root to do that
    context.sudo_user = context.get('sudo_user') or 'postgres'
    context.chunking = True  # needed by these commands

    replaces = sequence(section.pop('replace', None))
    if replaces:
        cfghandle({'replace': replaces}, cmds, context)

    # access
    perms = sequence(section.pop('access', None))
    if perms:
        cfghandle({'access': perms}, cmds, context)

    supwd = section.pop('su-password', None)
    if supwd:
        if supwd.startswith('md5 '):  # already md5d
            supwd = shplit(supwd)[1]
        else:
            supwd, errs  = find_password(supwd)
            errors += errs
            # http://www.postgresql.org/docs/9.1/static/catalog-pg-authid.html
            supwd = 'md5' + md5(supwd+'postgres')

        result = runcmd(cmds.pg_get_pwd, **context)
        if result.failed:
            log.error(result); errors += 1
        result = result.strip()
        log.debug('password: existing hash %r.', result)
        log.debug('password:   ...new hash %r.', supwd)

        if result == supwd:
            log.skip('password matches.')
        else:
            log.change('setting password.')
            result = runcmd(cmds.pg_set_pwd, supwd, **context)
            if result.succeeded:   changed += 1
            else:
                log.error(result); errors += 1
                if env.pave_raise_errs: raise RuntimeError

    for user in sequence(section.pop('user', [])):
        # look for existing
        result = runcmd(cmds.pg_ls_user, **context)
        if result.succeeded:  existing = result.split()
        else:
            log.error(result); errors += 1 # don't stop on search failure
            existing = []

        username = user['name']
        if username in existing:
            log.skip('user: %s found.', username)
        else:
            log.change('user: creating %s.', username)
            mycon = context.copy()
            mycon.update(user)
            result = runcmd(cmds.pg_cr_user, username, **mycon)
            if result.succeeded:   changed += 1
            else:
                log.error(result); errors += 1
                if env.pave_raise_errs: raise RuntimeError

    for db in sequence(section.pop('db', [])):
        # look for existing
        result = runcmd(cmds.pg_ls_db, **context)
        if result.succeeded:  existing = result.split()
        else:
            log.error(result); errors += 1 # don't stop on search failure
            existing = []

        dbname = db['name']
        if dbname in existing:
            log.skip('db: %s found.', dbname)
        else:
            log.change('db: creating %s.', dbname)
            mycon = context.copy()
            mycon.update(db)
            desc = db.get('description','')
            if desc: desc = u'"%s"' % desc
            result = runcmd(cmds.pg_cr_db, dbname, desc, **mycon)
            if result.succeeded:   changed += 1
            else:
                log.error(result); errors += 1
                if env.pave_raise_errs: raise RuntimeError

    for grant in sequence(section.pop('grant', [])):
        # look for existing
        privs, objtype, dbname, uname = grant.split()
        if objtype == 'db': objtype = 'database'
        # look for existing acl
        result = runcmd(cmds.pg_ls_acl, dbname, **context)
        if result.succeeded:
            acl = (len(result.split()) == 3)
        else:
            log.error(result); errors += 1 # don't stop on search failure
            acl = False

        if acl:
            log.skip('grant: acl for %s %s found.', dbname, objtype)
        else:
            log.change('grant: '+cmds.pg_grant.partition(' -c ')[2], privs,
                        objtype, dbname, uname)
            result = runcmd(cmds.pg_grant, privs, objtype, dbname, uname,
                            **context)
            if result.succeeded:   changed += 1
            else:
                log.error(result); errors += 1
                if env.pave_raise_errs: raise RuntimeError

    for statement in sequence(section.pop('sql', [])):
        # not sure how to skip this, maybe save a file.
        dbname = statement.split()[0]
        statement = statement[len(dbname) + 1:]
        log.change(u'sql: running %.20r… against %s.', statement, dbname)
                    #~ statement.encode('ascii','backslashreplace'), dbname)
        result = runcmd(cmds.pg_sql, dbname, statement, **context)
        if result.succeeded:   changed += 1
        else:
            log.error(result); errors += 1
            if env.pave_raise_errs: raise RuntimeError

    if section:
        log.warn('extra junk in section: %s', section.keys())

    return changed, errors














# (C) 2012-2013 Mike Miller.
# License: GNU GPLv3+
'''
    Create operating system user groups.

    Example::

        - groups:           # str || list || conditional
            - appgroup
            - "dev:1100"    # how to set gid

            - if-exec: test -e /dev/cdrom
              do: cdrom
'''
import logging

from fabric.api import env
from fabric.context_managers import hide
from voluptuous import Any

from pave.schema import conditionals_schema
from pave.utils import runcmd, sequence, check_list_conditionals

log = logging.getLogger()
schema = Any(basestring, [basestring, conditionals_schema])


def handle(section, cmds, context):
    changed, errors = 0, 0
    context.chunking = True  # needed by these commands

    # find existing:
    with hide('stdout'):  # useless text
        existing = runcmd(cmds.grp_lst)  # doesn't need to run as root
    if existing.failed:  errors += 1
    if existing:
        existing = existing.split()
    log.debug('existing groups: %s', existing)

    for group in sequence(section):
        group = check_list_conditionals(group, context, module=__name__)
        if not group: continue

        mycon = context.copy() # prevent context spill
        if ':' in group:
            group, mycon.gid = group.split(':', 1)

        if group in existing:
            log.skip('"%s" already exists.', group)
        else:
            log.change('creating group: %s', group)
            result = runcmd(cmds.grp_add, group, **mycon)

            if result.succeeded:   changed += 1
            else:
                log.error(result); errors += 1
                if env.pave_raise_errs: raise RuntimeError

    return changed, errors




# inspector - queries host details then caches the result in the temp folder.
import os, json, platform as p
from locale import getdefaultlocale as gdl
from os.path import join, expanduser

# work dir
wd = join(os.environ.get('XDG_CACHE_HOME') or expanduser('~/.cache'), 'pave')
# cache file name
cfn = join(wd, 'platform.json')

# grab the details
def query():
    details = dict(encoding=gdl()[1],
        arch=p.architecture(),machine=p.machine(),node=p.node(),
        proc=p.processor(),py_vers=p.python_version(),
        py_vert=p.python_version_tuple(),os_release=p.release(),
        system=p.system(),version=p.version(),win_vers=p.win32_ver(),
        mac_vers=p.mac_ver(),linux_dist=p.linux_distribution(),
    )
    return details

# check if cache file exists and > 0
if not (os.access(cfn, os.R_OK) and os.stat(cfn).st_size > 5):
# create the Cache file, using info from one of the files below.
    if not os.path.exists(wd):
        os.makedirs(wd)
        os.chmod(wd, 0750)
    with file(cfn, 'w') as f:
        json.dump(query(), f, separators=(',',':'))

# print results file to stdout
with file(cfn) as f:
    print f.read()



'''
    (C) 2012-2013 Mike Miller.
    License: GNU GPLv3+
    Where platform-specific commands and settings are kept.

    Class names should match the identifier string given by Python's
    platform module.  The version number can be appended as well to narrow its
    specificity.
'''
import re

deb_pkg_out_pat = re.compile(r'^(\d+) upgraded, (\d+) newly installed, ' +
                             r'(\d+) to remove')  # first three ints


class OS(object):
    python      = 'python "%s"'
    pythonc     = 'python2 -c "%s"'
    # No use to override __getattr__ etc unless using instances.
    # Currently using classes.
    @classmethod
    def _get_fqn(myclass):
        return str(myclass).split("'")[1]


class Windows(OS):
    pass


class Unix(OS):
    dir_test    = 'test -d "%s"'
    dir_create  = 'mkdir -p "%s"'
    file_append = '''echo -e '%s' >> "%s"'''
    file_chgrp  = 'chgrp %s "%s"'
    file_chmod  = 'chmod %s "%s"'
    file_chown  = 'chown %s "%s"'
    file_copy_q = 'cp "%s" "%s"'
    file_create = 'touch "%s"'
    file_exists = 'test -e "%s"'
    file_hash   = 'md5sum "%s"'
    file_search = '''grep -F '%s' "%s"'''
    file_test   = 'test -f "%s"'
    file_update = 'cp -f %s %s'
    home_path   = '/home'
    test_nt     = "test '%s' -nt '%s'"
    tempdir     = '/tmp'

    pg_ln_curr  = 'ln -sf "/etc/postgresql/%s" /etc/postgresql/curr'
    pg_get_pwd  = "psql -d template1 -t -c 'select passwd from pg_shadow;' "
    pg_set_pwd  = ('psql -c "ALTER USER postgres WITH ENCRYPTED PASSWORD'
                    ''' '%s';" template1 ''')
    pg_ls_user  = ("psql -d template1 -t -c 'select u.usename from "
                   "pg_catalog.pg_user u;' ")
    pg_cr_user  = ('createuser [--encrypted|no-encrypted] '
                   '[--createdb|no-createdb] [--createrole|no-createrole] '
                   '[--superuser|no-superuser] [--login|no-login] %s')
    pg_ls_db    = ("psql -d template1 -t -c 'SELECT datname FROM pg_database "
                   "WHERE datistemplate = false;' ")
    pg_cr_db    = ('createdb [--locale] [--encoding] [--tablespace] '
                   '[--template] %s %s')
    pg_ls_acl   = ('psql -d template1 -t -c "select datname, datacl from '
                   '''pg_database where datname = '%s';" ''')
    pg_grant    = ("psql -d template1 -c 'GRANT %s PRIVILEGES ON %s "
                   "%s to %s;' ")
    pg_sql      = 'psql -d %s -t -c "%s" '

    ssh_keypath = '${HOME}/.ssh/id_%s'
    ssh_akfile  = '${HOME}/.ssh/authorized_keys'
    ssh_keygen  = '''ssh-keygen -t %s -f "%s" -N '%s' '''
    svc_restart = 'service %s restart'

    dj_mancmd   = './manage.py'
    dj_chktable = (r"echo '\t\d' | ./manage.py dbshell | "
                    "/bin/grep -m 1 ' table ' ")
    dj_syncdb   = '%s syncdb --noinput --all' % dj_mancmd
    dj_chkmigrt = 'ls */migrations/*.pyc'
    dj_initmgrt = '%s migrate --fake --noinput' % dj_mancmd
    dj_loaddata = '%s loaddata %%s' % dj_mancmd
    # bash only
    #~ dj_chkdata  = '''
        #~ d=( $(/bin/grep -E '(pk|model)' %s|head -2|tr -d ',"'|tr . _|cut -f 3 -d:) )
        #~ echo search: ${d[@]}
        #~ echo "select * from ${d[1]} where id = $d;"|./manage.py dbshell|grep $d'''
    # py 2.6, perhaps awk would be better?
    dj_chkdata  = OS.pythonc % r'''
import sys, glob as g, commands as c
pk, tb = None, None
with file(sorted(g.glob('%s'))[0]) as f:
  for line in f:
    if pk and tb: break
    if 'pk' in line:
      pk = line.split()[1][:-1]
    elif 'model' in line:
      tb = line.split()[1][1:-2].replace('.', '_')
print 'search:', pk, tb
cmd = ('echo \'select id from %%s where id = %%s;\' | ./manage.py dbshell | grep %%s'
  %% (tb, pk, pk))
res = c.getoutput(cmd).strip()
print 'found :', res
if not res: sys.exit(1)'''


if True:

    class Linux(Unix):
        locale_dir      = '/usr/share/locale'
        locale_test     = 'find . -maxdepth 1 -type d | wc -l'
        locale_cln      = ("/bin/bash -O extglob -c 'rm -rf %s/!(%%s)/'"
                            % locale_dir)
        file_stat       = 'stat -c "%%a %%U %%G" "%s"'

        grp_add         = 'addgroup [--gid] %s'
        grp_lst         = 'cut -d: -f1 /etc/group'

        krn_get_param   = 'sysctl -n %s'
        krn_set_param   = 'sysctl -w "%s"'
        krn_param_fnm   = '/etc/sysctl.conf'
        krn_param_reg   = r'^\s*%s\s*=.*'  # ' vm.swappiness = ...'
        krn_srch_file   = "grep '%s' %%s" % krn_param_reg

        svc_lst         = 'service --status-all'
        svc_start       = 'service %s start'
        svc_stop        = 'service %s stop'
                          # change the name of the param with [--old new]
        usr_add         = ('useradd [--create-home] [--groups] ' +
                           '[--shell] [--password] %s')
        usr_lst         = 'cut -d: -f1 /etc/passwd'

        svc_disable     = 'update-rc.d %s disable'

        @staticmethod
        def parse_svclist(svclist):
            'Handles sysvinit format.'
            services = {}
            for service in svclist.split('\n'):
                tokens = service.split()
                services[tokens[3]] = (1  if tokens[1] == '+'  else 0)
            return services


    if True:

        class Debian(Linux):
            pkg_clean       = 'apt-get -y -q clean; apt-get -y -q autoremove'
            pkg_install     = 'DEBIAN_FRONTEND=noninteractive apt-get -y -q install %s'
            pkg_inventory   = "dpkg-query --show | cut -f1 | tr '\\n' ' ' "
            pkg_remove      = 'DEBIAN_FRONTEND=noninteractive apt-get -y -q purge %s'
            pkg_update      = 'apt-get -y -q update'
            pkg_update_time = 'find /var/lib/apt -name lists -daystart -mtime +%s'
            pkg_upgrade     = 'DEBIAN_FRONTEND=noninteractive apt-get -y -q upgrade'
            pkg_upgrd_full  = 'DEBIAN_FRONTEND=noninteractive apt-get -y -q dist-upgrade'
            pkg_upgrd_time  = 'find /var/lib/dpkg -name status -daystart -mtime +%s'
            pkg_stat_mtime  = 'stat -c %y /var/lib/dpkg/status'

            @staticmethod
            def parse_pkg_output(output):
                '0 upgraded, 0 newly installed, 0 to remove and 20 not upg...'
                if output:
                    for line in output.splitlines():
                        match = deb_pkg_out_pat.match(line)
                        if match:
                            return [ int(digit) for digit in match.groups() ]
                return []

        class Ubuntu(Debian):
            svc_lst         = 'service --status-all 2> /dev/null; initctl list'
            svc_disable_ups = ('mv -f /etc/init/%s.conf /etc/init/%s.disabled;'
                               ' echo "manual" > /etc/init/%s.override')

            @staticmethod
            def parse_svclist(svclist):
                'Handle sysvinit and upstart formats.'
                services = {}
                for service in svclist.split('\n'):
                    tokens = service.split()
                    if tokens[0] == '[':  # sysvinit
                        services[tokens[3]] = (1  if tokens[1] == '+'  else 0)
                    else:                 # upstart
                        status = (tokens[2] if tokens[1].startswith('(')
                                            else tokens[1])
                        status = status.partition('/')[0]
                        services[tokens[0]] = (1  if status == 'start'  else 0)
                return services


        class Raspbian(Debian):
            pass


        class Redhat(Linux):
            pkg_clean       = 'yum -y -q clean packages'
            pkg_install     = 'yum -y -q install %s'
            pkg_inventory   = 'rpm -qa | rev | cut -f 3- -d- | rev | sort'
            pkg_remove      = 'yum -y -q remove %s'
            pkg_update      = 'yum makecache'
            pkg_update_time = 'find /var/cache/yum/*/* -name updates -daystart -mtime +%s'
            pkg_upgrade     = 'yum -y -q update'
            pkg_upgrd_time  = 'find /var/lib -name yum -daystart -mtime +%s'
            pkg_stat_mtime  = 'stat -c %y /var/lib/yum'

            svc_start       = 'chkconfig %s on'
            svc_stop        = 'chkconfig %s off'

            @staticmethod
            def parse_svclist(svclist):
                'name (pid XXX) is running... # :/'
                services = {}
                for service in svclist.split('\n'):
                    tokens = service.split()
                    service[tokens[0]] = (1  if 'running...' in tokens  else 0)
                return services

        Red = Redhat # alias

        class CentOS(Redhat):
            pass

        class Fedora(Redhat):
            svc_lst         = 'systemctl list-unit-files -t service'

            @staticmethod
            def parse_svclist(svclist):
                services = {}
                for service in svclist.split('\n'):
                    tokens = service.split()
                    name = tokens[0].partition('.')[0]
                    service[name] = (1 if tokens[1] == 'enabled'  else 0)
                return services


        class Gentoo(Linux):
            grp_add         = 'groupadd [--gid] %s'
            pkg_clean       = 'eclean -q -C distfiles'
            pkg_install     = 'emerge -q --update --newuse %s'
            pkg_inventory   = "equery -q list --format='$name' '*'"
            pkg_remove      = 'emerge --quiet-unmerge-warn --unmerge %s'
            pkg_update      = 'emerge -q --sync'
            # update_time /var/db/pkg ?
            pkg_update_time = 'find /usr/portage/metadata -name "timestamp.x" -daystart -mtime %s'
            pkg_upgrade     = 'emerge -q --update --deep --newuse world'
            # upgrd_time - how to find out if a package was installed?
            # timestamps and/or by parsing the pkg mgr output.
            # see parse_pkg_output above
            pkg_upgrd_time  = 'find /var/tmp -name portage -daystart -mtime +%s'
            pkg_stat_mtime  = 'stat -c %y /var/tmp/portage'

            svc_lst         = 'rc-status -s -C'
            svc_start       = 'rc-service %s start'
            svc_stop        = 'rc-service %s stop'
            svc_disable     = 'rc-update del %s default'

            @staticmethod
            def parse_svclist(svclist):
                services = {}
                for service in svclist.split('\n'):
                    tokens = service.split()
                    services[tokens[0]] = (1 if tokens[2] == 'started' else 0)
                return services


    class Solaris(Unix):
        pass


    class BSD(Unix):
        pass


    if True:

        class Darwin(BSD): # OSX
            home_path   = '/Users'















# (C) 2012-2013 Mike Miller.
# License: GNU GPLv3+
'''
    Manage local/remote ssh keys and their generation.

    ``remote:``
        Generates a key at the remote host under the given username.

    ``copy-id:``
        Copies local keys to the remote host.

        Values may be a single or list of Boolean
        (True copies the public key of the local user to the authorized keys
        file of the remote user),
        or string of two arguments containing the local file
        (.pub appended automatically) and the remote user.

    Example::

        - keygen:
            remote: "%sitename"             # quotes not required but
            copy-id:                        # fixes syntax highlighting
                - True                          # default src dest
                - "%local_path.pub %user"       # specified

'''
import logging
from os.path import expandvars, expanduser, exists, dirname, join

from fabric.api import env
from fabric.operations import local
from voluptuous import Any, Required as req

from pave.schema import conditionals_schema
from pave.utils import runcmd, sequence, shplit, eval_conditionals, err_cond

# defaults
_def_type = 'rsa'
_def_psfr = ''

log = logging.getLogger()
schema = {
    req('copy-id', default=True):           Any(bool, basestring,
                                                [basestring, bool]),
    req('remote', default=None):            Any(None, basestring),
    req('type', default=_def_type):         basestring,
    req('passphrase', default=_def_psfr):   basestring,
}
schema.update(conditionals_schema)

def keygen(type=_def_type, passphrase=_def_psfr, **context):
    'Generate ssh keys locally.'
    from pave.main import lcmds
    locpath = expandvars(lcmds.ssh_keypath % type)

    # check just the dirname in case the user already has a different type key.
    if exists(dirname(locpath)):
        log.debug('Found local .ssh folder at "%s".', dirname(locpath))
    else:
        cmdstr = lcmds.ssh_keygen % ( type, locpath, passphrase)
        log.note('Generating local ssh keys with: %r', cmdstr)
        result = local(cmdstr, capture=True)
        if result.failed:
            log.error(result)
            if env.pave_raise_errs: raise RuntimeError

def check_ssh_config():
    'Check for config locally.'
    from pave.main import lcmds
    cfgpath = join(dirname(expandvars(lcmds.ssh_keypath)), 'config')
    if exists(cfgpath):
        log.debug('"%s" found, using ssh config.', cfgpath)
        env.use_ssh_config = True
    else:
        log.debug('"%s" not found, not using ssh config.', cfgpath)


def handle(section, cmds, context):
    changed, errors = 0, 0
    if eval_conditionals(section, context) is False:
        log.skip(err_cond, '')  # log here to tag from this module
        return changed, errors

    from pave.main import lcmds

    for copyid in sequence(section['copy-id']):
        if copyid:
            remakf = cmds.ssh_akfile
            remdir = dirname(remakf)
            mycon = context.copy()
            if copyid is True:
                locbase = expandvars(lcmds.ssh_keypath % section['type'])
                mycon.use_sudo, remusr = False, ''
            else:
                locbase, remusr = shplit(copyid)
                locbase = expanduser(locbase)
                mycon.use_sudo = True
                mycon.sudo_user = remusr
            # ensure we get the public key file
            locpath = (locbase if locbase.endswith('.pub') else locbase+'.pub')

            result = runcmd(cmds.dir_test, remdir, **mycon)
            if result.succeeded:
                log.skip('copy-id: .ssh folder already created. %s', remusr)
            else:
                log.change('copy-id: creating .ssh %s', remusr)
                result = runcmd('mkdir -p %s', remdir, **mycon)
                result = runcmd('chmod 700 %s', remdir, **mycon)
                if result.succeeded:   changed += 1
                else:
                    log.error(result); errors += 1
                    if env.pave_raise_errs: raise RuntimeError

            with file(locpath) as f:
                for line in f:
                    line = line.rstrip()

                    result = runcmd(cmds.file_search, line, remakf, **mycon)
                    if result.succeeded:
                        log.skip('copy-id: local key found at remote host.')
                    else:
                        log.change('copy-id: adding key to remote host.')
                        result = runcmd(cmds.file_append, line, remakf,**mycon)
                        if result.succeeded:  changed += 1
                        else:
                            log.error(result); errors += 1
                            if env.pave_raise_errs: raise RuntimeError

    for username in sequence(section['remote']):
        if username:
            mycon = context.copy()
            mycon.sudo_user = username
            mycon.use_sudo = True
            pkpath = cmds.ssh_keypath % section['type']
            result = runcmd(cmds.file_test, pkpath, **mycon)
            if result.succeeded:
                log.skip('remote: remote key found.')
            else:  # file doesn't exist
                log.change('remote: keygen for user %s.' % username)
                result = runcmd(cmds.ssh_keygen, section['type'], pkpath,
                                section['passphrase'], **mycon)
                if result.succeeded:  changed += 1
                else:
                    log.error(result); errors += 1
                    if env.pave_raise_errs: raise RuntimeError

    return changed, errors




# -*- coding: utf-8 -*-
# (C) 2012-2013 Mike Miller.
# License: GNU GPLv3+
'''
    Django helper module.
    *Captain Obvious recommends Django be installed on the target
    beforehand, see the* `packages`_ *section.*

    ``loaddata:``
        Load fixture files from given path, glob ok.

    ``manage:``
        Execute a management command every time this task group is run.

    ``manage-once:``
        Once and never again.

    Example::

        - django:
            workdir: ~%sitename/%sitename
            loaddata: "fixtures/*.json"
            manage-once: update_index
            syncdb: True
'''
import logging
from fabric.api import env
from fabric.context_managers import cd
from voluptuous import Any, Required as req

from pave.schema import conditionals_schema
from pave.utils import (runcmd, sequence, donepath, markdone,
                        eval_conditionals, err_cond)

log = logging.getLogger()
schema = {
    req('workdir'):                     basestring,
    req('syncdb', default=True):        bool,
    req('migrate-init', default=True):  bool,
    req('loaddata', default=''):        basestring,
    req('manage', default=[]):          Any(basestring, [basestring]),
    req('manage-once', default=[]):     Any(basestring, [basestring]),
    req('user', default=None):          basestring,
}
schema.update(conditionals_schema)


def handle(section, cmds, context):
    'Run the system package manager, and any others given.'
    changed, errors = 0, 0
    if eval_conditionals(section, context) is False:
        log.skip(err_cond, '')  # log here to tag from this module
        return changed, errors

    if section['user']:
        context.use_sudo = True
        context.sudo_user = section['user']

    with cd(section['workdir']):

        if section['syncdb']:
            result = runcmd(cmds.dj_chktable, **context)
            if result.succeeded:  # found a table
                log.skip(u'syncdb: already done.')
            else:
                log.change(u'syncdb: may take a few mins…')
                result = runcmd(cmds.dj_syncdb, **context)
                if result.succeeded:   changed += 1
                else:
                    log.error(result); errors += 1
                    if env.pave_raise_errs: raise RuntimeError

        if section['migrate-init']:
            result = runcmd(cmds.dj_chkmigrt, **context)
            if result.succeeded:
                log.skip(u'initial migrate has already been run.')
            else:
                log.change(u'initial migration…')
                result = runcmd(cmds.dj_initmgrt, **context)
                if result.succeeded:   changed += 1
                else:
                    log.error(result); errors += 1
                    if env.pave_raise_errs: raise RuntimeError

        if section['loaddata']:
            # check is not yet implemented, perhaps parse the json and
            # confirm an id in the table.
            result = runcmd(cmds.dj_chkdata % section['loaddata'],
                            chunking=False, **context)
            if result.succeeded:
                log.skip(u'loaddata: fixtures already loaded.')
            else:
                log.change(u'loaddata: fixtures…')
                result = runcmd(cmds.dj_loaddata, section['loaddata'],
                                **context)
                if result.succeeded:   changed += 1
                else:
                    log.error(result); errors += 1
                    if env.pave_raise_errs: raise RuntimeError

        for manparam in sequence(section['manage']):
            log.change(u'manage: %s…' % manparam)
            result = runcmd('%s %s' % (cmds.dj_mancmd, manparam), **context)
            if result.succeeded:   changed += 1
            else:
                log.error(result); errors += 1
                if env.pave_raise_errs: raise RuntimeError

        for manparam in sequence(section['manage-once']):
            cmd = manparam.split()[0]
            tdir, fpath = donepath(cmds, 'done.django.manage-once', cmd, manparam)
            # test file
            result = runcmd(cmds.file_test, fpath, **context)
            if result.succeeded:
                log.skip(u'manage-once: %s… has been run already.', cmd)
            else:
                log.change(u'manage-once: %s…' % manparam[:30])
                result = runcmd('%s %s' % (cmds.dj_mancmd, manparam), **context)
                if result.succeeded:
                    changed += 1
                    errors += markdone(cmds, tdir, fpath, context)
                else:
                    log.error(result); errors += 1
                    if env.pave_raise_errs: raise RuntimeError

    return changed, errors  # only one change




# -*- coding: utf-8 -*-
'''
    (C) 2012-2013 Mike Miller.
    License: GNU GPLv3+

    A grab bag of assorted stuff.
'''
import sys, logging, hashlib, os, re, random, string
from os.path import join
from getpass import getpass
from pipes import quote  # use shlex in py 3.3+
from platform import python_version_tuple

from yaml import SafeLoader
from fabric.api import env, local, run, sudo
from fabric.contrib.files import exists as rexists, _escape_for_regex

from meta import __progname
import platforms

# shlex & unicode :/
if python_version_tuple() < ('2', '7', '3'):
    from pave.uni_shlex import split as shplit
else:
    from shlex import split as shplit
# wait key
try:                    # Win32
    from msvcrt import getch as _getch
except ImportError:     # UNIX
    def _getch():
        import tty, termios
        fd = sys.stdin.fileno()
        old = termios.tcgetattr(fd)
        try:
            tty.setraw(fd)
            return sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old)

log = logging.getLogger()
CURRENT = 'CURRENT'
shell_error_map = {  # from bash docs
    1: 'General Error',
    2: 'Misuse of shell builtins',
    126: 'Command invoked cannot execute.',
    127: 'Command not found.',
    128: 'Invalid argument to exit',
    130: 'Script terminated by Control-C',
    255: 'Exit status out of range',
}
err_cond = 'test(s) %sreturned a False.'


class PaveSafeLoader(SafeLoader):
    ''' Disables the use of "%" chars as directive characters, so we don't have
        to quote so much. '''
    def check_directive(self):
        return False
    def check_plain(self):
        # removes % from the list
        ch = self.peek()
        return ch not in u'\0 \t\r\n\x85\u2028\u2029-?:,[]{}#&*!|>\'\"@`'  \
                or (self.peek(1) not in u'\0 \t\r\n\x85\u2028\u2029'
                    and (ch == u'-' or (not self.flow_level and ch in u'?:')))


class AttrDict(dict):
    'Dict that acts like an object.'
    def __getattr__(self, attr):
        return self[attr]

    def __setattr__(self, attr, value):
        self[attr] = value

    def copy(self):
        return AttrDict(self)

# allow unicode var names in templating:
class _VarTemplateMetaclass(string._TemplateMetaclass):
    def __init__(cls, name, bases, dct):
        super(_VarTemplateMetaclass, cls).__init__(name, bases, dct)
        if 'pattern' in dct:
            pattern = cls.pattern
        else:
            pattern = _VarTemplateMetaclass.pattern % {
                'delim' : re.escape(cls.delimiter),
                'id'    : cls.idpattern }
        cls.pattern = re.compile(pattern, re.IGNORECASE|re.VERBOSE|re.UNICODE)

class VarTemplate(string.Template):
    __metaclass__ = _VarTemplateMetaclass
    delimiter = '%'
    idpattern = r'[_\w][_\w0-9]*'


def backup(fname, cmds, context):
    'Make a .orig backup of files we are going to modify.'
    bakname = fname + '.orig'
    if not rexists(bakname):
        result = runcmd(cmds.file_copy_q, fname, bakname, **context)
        if result.failed:
            log.warn(result)
            # if env.pave_raise_errs: raise RuntimeError # don't quit
        else:
            log.debug('backed up "%s" to .orig', fname)


def bold(text):
    'Add a bold attribute to given text.'
    return '\033[1m%s\033[0m' % text


# see logcfg for custom logger to change module name
def check_list_conditionals(item, context, module=None, replace=True,
                            force=False):
    ''' Check if item is a dictionary and if so eval conditionals.
        Helpful For list-based modules.
        Returns None if item should be skipped.
        Replace - replace object with its do/then clause.
    '''
    if type(item) is dict:
        if module:  extra = {'module': module.split('.')[-1]}
        else:       extra = None

        thenstr = item.get('then') or item.get('do')
        title = item.get('title')
        if force or thenstr:
            if eval_conditionals(item, context) is False:
                log.skip(err_cond, u'for "%.30s…" ' % (title or thenstr),
                         extra=extra)
                item = None
            elif replace:
                item = thenstr
        elif replace:           # this should be handled in schema
            log.error('conditional missing "then/do:" key.')
            raise RuntimeError
    return item


def contains(filename, text, exact=False, use_sudo=False, escape=False,
             fixed=False):
    ''' Copied from fabric--shows output and escape defaults to False.
        I may remove this and reimplement.
    '''
    func = use_sudo and sudo or run
    if escape:
        if not fixed:
            text = _escape_for_regex(text)
        if exact:
            text = '^%s$' % text
    else:
        text = text.replace("'", "\'")
    if fixed: fixed = '-F'
    else:     fixed = '-E'
    egrep_cmd = '''grep %s -- '%s' "%s" ''' % (fixed, text, filename)
    result = func(egrep_cmd, shell=False)
    log.debug('return code: %s', result.return_code)
    return result.succeeded


# crypt ----------------------------------------------------------------
def ask_crypt(tries=3):
    '''Asks for a password and returns crypt'd.'''
    for attempt in range(tries):
        passwd  = getpass('Enter password to crypt: ')
        passwd2 = getpass('Enter again to confirm:  ')
        if passwd == passwd2:
            if passwd:
                return crypt(passwd)
            else:
                log.warn('Passwords empty, exiting.')
                return
        else:
            log.error('Passwords do not match, try again.')
    log.critical('Fail.')


def crypt(passwd):
    'Crypt a password. See ``man crypt`` for details.'
    try:
        from crypt import crypt as _crypt
    except ImportError:
        try:
            from fcrypt import crypt as _crypt
            print '\nWarning: using non-native crypt support (fcrypt).'
        except ImportError:
            print '\nError: crypt support not available.  Install fcrypt:'
            print '       pip install fcrypt\n'

    salt = [
        chr( random.randrange(1, 27) + random.choice((64,96)) ),
        chr( random.randrange(1, 27) + random.choice((64,96)) ),
        chr( random.randrange(10) + 48 ),
        chr( random.randrange(10) + 48 ),
    ]
    random.shuffle(salt)
    salt = ''.join(salt)
    encrypted = _crypt(passwd, '$5$' + salt)
    if log.isEnabledFor(logging.DEBUG):
        #~ log.debug('crypt.passwd: %r', passwd.encode('base64').rstrip())
        log.debug('crypt.salt: %r', salt)
        log.debug('crypt.result: %r', encrypted)
    return encrypted
# crypt ----------------------------------------------------------------

# done -----------------------------------------------------------------
def donepath(cmds, base, itemname, item):
    'Create and return a path to a "done" file.'
    path = base + '.%s.%s' % (itemname, md5(item))
    tdir = join(cmds.tempdir, __progname + '-$USER')
    fullpath = join(tdir, path)
    return tdir, fullpath


def markdone(cmds, tdir, path, context):
    'Mark a task as "done".'
    locerrs = 0
    result = runcmd(cmds.dir_create, tdir, **context)
    if result.failed:
        log.error(result); locerrs += 1
        if env.pave_raise_errs: raise RuntimeError
    result = runcmd(cmds.file_create, path, **context)
    if result.failed:
        log.error(result); locerrs += 1
        if env.pave_raise_errs: raise RuntimeError
    return locerrs
# done -----------------------------------------------------------------


def eval_conditionals(sectiondict, context):
    'If multiple tests, must all be true; boolean AND.'
    vartest = sectiondict.pop('if', None)
    exectest = sectiondict.pop('if-exec', None)
    plattest = sectiondict.pop('if-platform', None)

    if vartest:
        val1, comp, val2 = shplit(vartest)  # schema can check length
        val1 = unicode(env.pave_vars.get(val1, val1))
        vartest = _strcmp(val1, comp, val2)
        log.debug(u'var-test: %s » %s', (val1, comp, val2), vartest)

    if exectest:
        exectest = runcmd(exectest, **context).succeeded
        log.debug(u'exec-test: » %s', exectest)

    if plattest:
        tokens = shplit(plattest)
        if len(tokens) == 1:
            theclass = getattr(platforms, tokens[0], bool) # bool def as false
            plattest = issubclass(env.pave_platform, theclass)
            log.debug(u'platform-test: %s ⊇ %s » %s', env.pave_platform._get_fqn(),
                      theclass._get_fqn(), plattest)
        else:
            details = env.pave_platform_details.get(env.host_string, {})
            val1, comp, val2 = tokens
            val1 = unicode(safe_eval(val1, **details))
            plattest = _strcmp(val1, comp, val2)
            log.debug(u'platform-test: %s » %s', (val1, comp, val2), plattest)

    tests = [ t for t in (vartest, exectest, plattest) if t is not None ]
    return all(tests)


def fmtcmd(cmd, *args, **options):
    'format a command line given both arguments and options.'
    result = ''
    chunks = re.split(r'[\[\]]', cmd)
    log.debug('chunks: %s', chunks)
    current_arg = 0
    for i, chunk in enumerate(chunks):
        if chunk.startswith('-'):   # is a kwarg
            subchunks = re.split(r'[ =]', chunk)
            if len(subchunks) > 1:
                argname = subchunks[0]
                name = subchunks[1]
            else:
                argname = subchunks[0]
                name = subchunks[0].lstrip('-')

            name, _, negative = name.partition('|')  # postgres :/
            prefix = ('--' if argname.startswith('--') else '-')

            if name in options:
                value = options[name]
                if value is None:
                    continue
                elif type(value) is bool:
                    if value:           chunk = argname.partition('|')[0]
                    else:
                        if negative:    chunk = prefix + negative
                        else:           continue
                else:
                    chunk = argname + ' ' + quote(value)
                result += chunk
        elif chunk.isspace():
            result += chunk
        elif chunk:                       # is an arg
            try:
                count = chunk.replace('%%', '').count('%')
                chunk = chunk % args[current_arg:count]
                current_arg += count
            except (IndexError, TypeError), e:
                log.error('%s: %s', e.__class__.__name__, e)
            result += chunk

    return result


def get_cursor_pos():
    ''' Return the current column number of the terminal cursor.
        Used to figure out if we need to print an extra newline.
    '''
    value = 2
    if sys.stdout.isatty():
        import string
        sys.stdout.write('\033[6n')
        response = ''
        while True:
            c = sys.stdin.read(1)
            if c == 'R': break
            elif (c in string.digits) or (c == ';'):
                response += c
        if response:
            response = response.partition(';')[2]
            if response:
                value = int(response)
    return value


def getpwd(prompt, tries=3):
    'Request a password until two copies match.'
    while tries:
        first = getpass(prompt + ': ')
        second = getpass(prompt + ' again:')

        if first == second:
            if first:
                log.info('Passwords match.')
            return first  # '' signals exit
        else:
            tries = tries - 1
            if tries: log.error('Passwords don\'t match!  Try again.')


def find_password(directive, encrypt=None):
    result = ''
    errors = 0

    if directive.startswith('use '):
        try:
            directive = shplit(directive)[1]
        except IndexError:
            log.error('incorrect password use-directive.')
            errors += 1
            if env.pave_raise_errs:     raise RuntimeError
        try:
            result = env.pave_passwords[directive]
        except KeyError:
            log.error('password "%s" not found.', directive)
            errors += 1
            raise RuntimeError
    elif directive:
        result = directive

    if result and encrypt:
        result = crypt(result)
    return result, errors


def md5(data):
    'Given a piece of data, return its md5 hash.'
    digest = hashlib.md5()
    digest.update(data)
    return digest.hexdigest()


def q(fname):
    'Quote strings (e.g. filenames) if needed.'
    if ' ' in fname:
        fname = u'"%s"' % fname
    return fname


def runcmd(cmdline, *args, **options):
    'Run a command with fabric, using arguments to decide on sudo or not.'
    use_sudo = options.pop('use_sudo', None)
    sudo_user = options.pop('sudo_user', None)
    is_local = options.pop('local', None)
    chunking = options.pop('chunking', None)

    if chunking and '[' in cmdline:
        cmdline = fmtcmd(cmdline, *args, **options)
    elif args:
        cmdline = cmdline % args  # allow formatting w/ chunking off

    if is_local:
        result = local(cmdline, capture=True)
    elif use_sudo:
        result = sudo(cmdline, user=sudo_user)
    else:
        result = run(cmdline)
    log.debug('return code: %s', result.return_code)
    return result


def safe_eval(text, **namespace):
    ''' Disables builtins and provides only given variables.
        http://lybniz2.sourceforge.net/safeeval.html
    '''
    return eval(text, {'__builtins__': None}, namespace)


def sequence(value, *toappend):
    ''' Given an object, check if it is a sequence.  If not, add it to one.
        Return the sequence.
    '''
    if type(value) not in (list, tuple):
        value = [value]
    if toappend:    # is a tuple
        if type(value) is list:
            toappend = list(toappend)
        value = value + toappend
    return value


def _strcmp(val1, comp, val2):
    'Simple string compare function.'
    result = None
    if comp == '==':
        if val1 == val2:    result = True
        else:               result = False
    elif comp == '!=':
        if val1 != val2:    result = True
        else:               result = False
    return result


def trunc(msg, length):
    'Truncate a string with trailing ellipsis.'
    if len(msg) > length:
        msg = msg[:length] + u'…'
    return msg


def wait_key():
    'Waits for a keypress at the console.'
    return _getch()


# add constants for Windows :/
try:
    os.EX_OK
except AttributeError:
    os.EX_OK = 0
    os.EX_USAGE = 64
    os.EX_DATAERR = 65
    os.EX_NOINPUT = 66
    os.EX_NOUSER = 67
    os.EX_NOHOST = 68
    os.EX_UNAVAILABLE = 69
    os.EX_SOFTWARE = 70
    os.EX_OSERR = 71
    os.EX_OSFILE = 72
    os.EX_CANTCREAT = 73
    os.EX_IOERR = 74
    os.EX_TEMPFAIL = 75
    os.EX_PROTOCOL = 76
    os.EX_NOPERM = 77
    os.EX_CONFIG = 78
    os.EX_NOTFOUND = 79




# -*- coding: utf-8 -*-
# (C) 2012-2013 Mike Miller.
# License: GNU GPLv3+
'''
    Good hygene is important.

    ``folders:``
        Deletion of files under ``/tmp``, ``/var/tmp``, & ``/home/`` only.
    ``locales-except:``
        *"Unless you're working at the UN and administer a central server for
        all member states, it is difficult to conceive why you would
        need a system where all of these locales are installed."* |br|
        Similar to a one-time Debian locale-purge.
    ``packages:``
        Run appropriate package cleanup command(s).

    Example::

        - cleanup:
            packages: True
            locales-except: en en_US
            folders:  # use with caution
                - /tmp/stuff/*
                - /tmp/stuff2

'''
import logging

from fabric.api import env
from fabric.context_managers import cd
from fabric.contrib.files import exists as rexists
from voluptuous import Required as req

try:  from pave.lib import packages  # circular at schema import, mv to handle?
except ImportError: pass
from pave.utils import runcmd, eval_conditionals, err_cond
from pave.schema import conditionals_schema

log = logging.getLogger()
whitelist = ('/tmp/', '/var/tmp/', '/home/')
schema = {
    req('folders', default=[]):         [basestring],
    req('packages', default=True):      bool,
    'locales-except':                   basestring,
}
schema.update(conditionals_schema)

def handle(section, cmds, context):
    changed, errors = 0, 0
    if eval_conditionals(section, context) is False:
        log.skip(err_cond, '')  # log here to tag from this module
        return changed, errors

    for task in section:
        if task == 'packages':
            if section['packages']:
                chd, errs, mods = packages.cleanup(cmds, context)
                if chd:
                    log.change(u'packages: removed %s obsolete.', mods)
                elif not errs:
                    log.skip(u'packages: none to remove.')
                changed += chd; errors += errs

        elif task == 'locales-except':
            with cd(cmds.locale_dir):
                # check to see what is there now:
                before = runcmd(cmds.locale_test, **context)
                if before.failed:
                    log.error(before); errors += 1
                    if env.pave_raise_errs: raise RuntimeError

                log.info(u'removing any extra locale files…')
                filespec = '|'.join(section[task].split())
                result = runcmd(cmds.locale_cln, filespec, **context)
                if result.succeeded:   pass  # changed += 1 # need to check if done
                else:
                    log.error(result); errors += 1
                    if env.pave_raise_errs: raise RuntimeError

                # what's it look like afterward?
                after = runcmd(cmds.locale_test, **context)
                if after.succeeded:
                    if before == after:
                        log.skip('locales: before %s, after %s', before, after)
                    else:
                        log.change('locales: before %s, after %s', before, after)
                        changed +=1
                else:
                    log.error(after); errors += 1
                    if env.pave_raise_errs: raise RuntimeError


        elif task == 'folders':
            for folder in section[task]:
                if '../' not in folder:
                    permitted = False
                    for path in whitelist:  #  check
                        if folder.startswith(path):
                            permitted = True
                            break
                    if permitted:
                        if rexists(folder):
                            log.change('removing %s', folder)
                            result = runcmd('rm -rf %s', folder, **context)
                            if result.succeeded:   changed += 1
                            else:
                                log.error(result); errors += 1
                                if env.pave_raise_errs: raise RuntimeError
                        else:
                            log.skip('"%s" is already deleted.', folder)
                    else:
                        log.error('deletion of "%s" not permitted.', folder)
                        errors += 1
                        if env.pave_raise_errs: raise RuntimeError
        else: # should be done by schema
            log.error('unknown cleanup task "%s"', task);  errors += 1
            if env.pave_raise_errs: raise RuntimeError

    return changed, errors




''' (C) 2012-2013 Mike Miller.
    License: GNU GPLv3+
'''

# need a module list to load schemas, etc
from os import listdir
from os.path import basename, dirname, splitext

modlist = set([
    splitext(basename(fn))[0]
    for fn in listdir(dirname(__file__))
    if fn.endswith('.py')
    if not basename(fn).startswith('_')
])




# (C) 2012-2013 Mike Miller.
# License: GNU GPLv3+
'''
    Manage kernel parameters.  Saves to memory and disk.

    Example::

        - kernel:           # str || list || conditional
            - vm.swappiness = 10

'''
import logging

from fabric.api import env
from fabric.contrib.files import sed
from voluptuous import Any

from pave.schema import conditionals_schema
from pave.utils import runcmd, sequence, check_list_conditionals

log = logging.getLogger()
schema = Any(basestring, [basestring, conditionals_schema])


def handle(section, cmds, context):
    changed, errors = 0, 0

    for variable in sequence(section):
        thischanged = 0
        variable = check_list_conditionals(variable, context, module=__name__)
        if not variable: continue

        name, value = [ arg.strip() for arg in variable.split('=', 1) ]

        # get current param value
        inmemory = runcmd(cmds.krn_get_param % name)
        if inmemory.failed: errors += 1
        log.debug('current mem  value: %s = %s', name, inmemory)

        # needs to be changed?
        if inmemory == value:
            log.skip('current mem  value matches: %s=%s', name, inmemory)
        else:
            log.change('setting mem  value: %s to %s', name, value)
            result = runcmd(cmds.krn_set_param % (name + '=' + value),
                            **context)
            if result.succeeded: thischanged += 1
            else:
                log.error(result); errors += 1
                if env.pave_raise_errs: raise RuntimeError

        # get current value on-disk
        # grep found 0 succeeded, not found 1 failed
        ondisk = runcmd(cmds.krn_srch_file, name, cmds.krn_param_fnm)
        if ondisk.succeeded:  # found
            ondiskval = ondisk.split('\n')[-1].partition('=')[2].strip()
            log.debug('current disk value: %s = %s', name, ondiskval)
        else:
            ondiskval = ''
            log.debug('disk value not found: %s', name)

        if ondiskval == value:
            log.skip('current disk value matches: %s=%s', name, ondiskval)
        else:
            log.change('setting disk value: %s to %s', name, value)
            if ondisk.succeeded:  # this greps the file again, not efficient:
                # was found, needs to be changed
                result = sed(cmds.krn_param_fnm, cmds.krn_param_reg % name,
                             variable, use_sudo=context.get('use_sudo'))
                if result.succeeded:    thischanged += 1
                else:
                    log.error(result); errors += 1
                    if env.pave_raise_errs: raise RuntimeError
            else:
                # not in file, add
                # append needs to be reimplemented, not enough info avail
                result = runcmd(cmds.file_append, variable, cmds.krn_param_fnm,
                                **context)
                if result.succeeded:  thischanged += 1
                else:
                    log.error(result); errors += 1
                    if env.pave_raise_errs: raise RuntimeError

        if thischanged:  # send one change per setting only
            changed += 1

    return changed, errors




# -*- coding: utf-8 -*-
'''
    %%prog v%s - (C) 2012-2013 Mike Miller.
    License: GNU GPLv3+
    Yet another config and deployment tool, leveraging fabric.

    %%prog [options]
'''
if True: # foldable
    import sys, os
    import copy, errno, logging, json, platform, string, traceback
    from os.path import dirname, join, expanduser
    from importlib import import_module
    from optparse import OptionParser
    from pprint import pformat

    import yaml
    from fabric.api import env, local, run, settings
    from fabric.exceptions import NetworkError
    from fabric.context_managers import hide, show
    from fabric.network import disconnect_all
    from fabric.tasks import execute
    from fabric.version import __version__ as fabver
    from voluptuous import Invalid as InvalidData

    from meta import __version__, __progname
    from logcfg import (console, ColorFmtr, EspeakHandler, ioLogger, log,
                        logfmt, start_file_log, SecshFilterD)

    _def_filename    = 'pave.yml'
    _def_encoding   = 'utf8'
    def_jobs        = 1
    pw_tries        = 3
    allvars         = {}
    opts            = None
    lcmds           = None

    try: # find a logging location
        import xdg.BaseDirectory  # modern linux
        tempdir = join(xdg.BaseDirectory.xdg_cache_home, __progname)
    except ImportError:
        xdg = None
        tempdir = join(expanduser('~'), '.cache', __progname)
    logdir = join(tempdir, 'logs')

    # some of these dependent on the values above
    import platforms
    from lib.keygen import keygen, check_ssh_config
    from schema import schema
    from targets import detect_range, expand_hostname_range
    from utils import ( AttrDict, ask_crypt, bold, getpwd, shplit, CURRENT,
                        get_cursor_pos, PaveSafeLoader, sequence, wait_key,
                        shell_error_map, safe_eval, VarTemplate )


def _expandstr(template, variables): # reusable
    return VarTemplate(template).substitute(variables)

def expandstr(self, node):   # expands vars
    'Yaml loader/constructor that expands variable strings.'
    if opts.brace_exp:
        if '{' in node.value and '}' in node.value:
            node.value = node.value.format(**allvars)
    else:
        if '%' in node.value:
            node.value = _expandstr(node.value, allvars)

    return self.construct_scalar(node)


def expand_targets(targets, groups):
    'Convert group names into targets and add them to the list.'
    result = []
    for target in targets:
        if target in groups:
            target = groups[target]
        if isinstance(target, list):
            result.extend(expand_targets(target, groups))
        else:
            match = detect_range(target)
            if match:   result.extend(expand_hostname_range(match))
            else:       result.append(target)
    return result


def fatal_exit(msg, exitcode, parser=None):
    'Convenience function to log and exit on fatal error.'
    if parser:  parser.print_help(); print
    log.fatal(msg); print
    logging.shutdown()
    sys.exit(exitcode)


def _fmt_log(shell, code):
    msg = 'inpsector %s exited with %s - %s'
    msg = msg % (shell.split()[0], code, shell_error_map.get(code, 'Unknown'))
    return msg


def get(cfgdata, path, offset=0):
    ''' An xpath-like yaml data grabber.
        Looks at these in order, last one wins:

            - Schema default
            - Config file
            - Command line
    '''
    # check command line, should be done better
    if path in opts.option:
        return opts.option[path]
    if path == '/main/targets' and opts.target:
        return opts.target
    if path == '/main/jobs' and opts.jobs > 1:
        return opts.jobs

    # check config file
    result = None
    rescfg = cfgdata
    names = path.split('/')[offset:]
    length = len(names)
    for i, name in enumerate(names):
        if name:
            if type(rescfg) is dict:
                if i == (length - 1):
                    rescfg = rescfg.get(name, None)
                else:
                    rescfg = rescfg.get(name, {})
            elif type(rescfg) is list:
                if rescfg:
                    if name == '*':
                        rescfg = tuple([ get(res, path, i+1) for res in rescfg ])
                    elif name.isdigit():
                        index = int(name)
                        if index < len(rescfg):
                            rescfg = rescfg[index]

    if rescfg is not None: # not in (None, {}, []): ?
        result = rescfg

    return result


def get_module(modname):
    '''Given a module name, find and load it.'''
    try:
        module = import_module('pave.lib.%s' % modname)
    except ImportError:     # look for third party modules
        module = import_module(modname)
    return module


def get_platform_cmds(pldata, local=False):
    'Given platform data, return the appropriate command class.'
    cmdobj = None
    sysname, vers = 'Linux', 'unknown'

    if pldata:
        # this doesn't happen early enough to make a difference:
        #~ allvars.update(pldata)
        if pldata.get('system') == 'Linux':
            distinfo = pldata.get('linux_dist')
            if len(distinfo) and distinfo[0]:  # list of 3 str
                sysname = distinfo[0].partition(' ')[0]  # cut from first space
                sysname = sysname.title()                # debian, grr...
            if len(distinfo) > 1 and distinfo[1]:
                vers = distinfo[1]
        else:
            sysname = pldata.get('system')
    if local:  log.debug('using a local %s system.' % sysname)
    else:      log.info('found a(n) %s system, version %s.' % (sysname, vers))

    # retrieve class
    vers = ''.join([ c  for c in vers  if c.isdigit() ])  # clean up
    classname = sysname + vers
    while True:
        try:  cmdobj = getattr(platforms, classname)
        except AttributeError:
            if classname and classname[-1].isdigit():
                classname = classname[:-1]  # chomp and try again
                continue
        break

    if cmdobj is None:
        fatal_exit('"%s/%s" is not yet supported.' % (pldata.get('system'),
                   ' '.join(pldata.get('linux_dist', []))), os.EX_UNAVAILABLE)
    return cmdobj


def load_data(opts, args):
    ''' Load the pavefile and expand any variables found.
    '''
    filedat, data = '', {}
    try:  # read in pavefile, first parse
        with file(opts.filename) as f:
            filedat = f.read()  # cache for later
            data = yaml.load(filedat, Loader=PaveSafeLoader)
        if not data:
            fatal_exit('pavefile "%s" empty.' % opts.filename, os.EX_DATAERR)

        # look at passwords early, so they can be expandable variables
        # look for them in includes
        includes = data.get('main', {}).get('include', [])
        includedat = ''
        if isinstance(includes, basestring):
            includes = shplit(unicode(includes))
        for incfname in includes:
            with file(incfname) as f:
                incfdat = f.read()
                pwds = yaml.load(incfdat,
                                 Loader=PaveSafeLoader).get('passwords', [])
                if pwds: data.update({'passwords': pwds})
                includedat += incfdat + '\n'  # cache for later

        # validate passwords section
        plines = schema({'main': {}, 'passwords': data.get('passwords', [])})
        plines = plines.get('passwords', [])
        # parse passwords section
        env.pave_passwords = {}
        for i, item in enumerate(plines):
            if isinstance(item, basestring):
                pargs = shplit(unicode(item))
                cmd, name = pargs[0], pargs[1]
                if cmd == 'ask':
                    prompt = ' '.join(pargs[2:])
                    passwd = getpwd(prompt, tries=pw_tries)
                    if not passwd:
                        raise EOFError, 'Password empty, exiting.'
                elif cmd == 'set':
                    passwd = pargs[2]
                env.pave_passwords[name] = passwd   # save to fabric's env
                allvars['pwd_' + name] = passwd     # to vars as well
                if name == 'password': env.password = passwd
            elif type(item) is dict:
                env.passwords = item
                env.pave_passwords = item

        # expand vars in vars section once
        allvars.update(data.get('vars', {}))
        for key, val in allvars.iteritems():
            if isinstance(val, basestring):
                if opts.brace_exp:
                    if '{' in val:   allvars[key] = val.format(**allvars)
                else:
                    if '%' in val:  allvars[key] = _expandstr(val, allvars)
        # command line override?
        for var in opts.var:
            allvars[var[0]] = var[1]
        env.pave_vars = allvars

        # look at pavefile a second time with var expander loaded
        if allvars:
            PaveSafeLoader.add_constructor(u'tag:yaml.org,2002:str', expandstr)
            # is it a template?
            if opts.filename.lower().endswith('.j2'):
                from jinja2 import Template
                template = Template(filedat)
                filedat = template.render(**allvars)
            data = yaml.load(filedat, Loader=PaveSafeLoader)

        # handle includes again
        if includedat:
            data.update(yaml.load(includedat, Loader=PaveSafeLoader))

        # rm password section from logging below; pwds attached to env obj
        for pline in data.get('passwords', []):
            if pline.startswith('set '):
                data.pop('passwords', None); break

        # do the full schema check
        data = schema(data)

        # printen Sie
        log.debug('schema:\n' + pformat(schema.schema))
        log.debug('pavefile:\n' + pformat(data))
        return data

    except EOFError, e:
        log.warn(str(e))
        sys.exit(os.EX_TEMPFAIL)
    except KeyboardInterrupt, e:
        print;  log.warn('Keyboard interrupt, exiting.')
        sys.exit(os.EX_TEMPFAIL)
    except InvalidData, e:
        fatal_exit('Invalid Data: %s - %s' % (e.msg,
                   '/' + '/'.join([unicode(p) for p in e.path])),os.EX_DATAERR)
    except KeyError, e:
        fatal_exit('pavefile: variable "%s" doesn\'t exist.' % e.args,
                    os.EX_DATAERR)
    except yaml.YAMLError, e:
        msg = ' '.join(unicode(e).split()) # normalize whitespace
        fatal_exit('yaml.%s: %s' % (e.__class__.__name__, msg), os.EX_DATAERR)
    except ImportError:
        fatal_exit('using jinja2 pavefile but jinja not found.', os.EX_CONFIG)
    except Exception, e:
        log.exc('%s:\n  %s', e.__class__.__name__, traceback.format_exc())
        fatal_exit('Exiting.', os.EX_SOFTWARE)


def main(opts, args):
    'Start paving.'
    exit_status = os.EX_OK
    # read the plans
    data = load_data(opts, args)

    # start your engines
    data, inspectors = prep_environ(data, opts)
    if opts.test:  return  # quit early
    results = {}
    context = AttrDict( use_sudo=get(data, '/main/sudo'),
                        sudo_user=get(data, '/main/sudo-user') )
    orig_stdout, orig_stderr = sys.stdout, sys.stderr # save for later
    sys.stdout = ioLogger(log, 'debug')               # redirect stdout to logs
    sys.stderr = ioLogger(log, 'debug')               # redirect stderr to logs
    everything = ('aborts', 'debug', 'running', 'status', 'stderr', 'stdout',
                    'user', 'warnings')  # fabric everything not everything :/
    log.debug(u'starting fabric.execute() on steamroll…')
    locfhandler = log.handlers.pop()  # don't double log to localhost

    # experimental use of tasks from external fabfile
    if 'fab-tasks' in data:  # start fabric
        from fabric import state
        from fabric.main import find_fabfile, load_fabfile, parse_arguments

        fabfile = find_fabfile()
        if not fabfile:
            fatal_exit('Couldn\'t find any fabfiles!', os.EX_IOERR)
        _, callables, _ = load_fabfile(fabfile)
        state.commands.update(callables)

        for i, fabtask in enumerate(data.get('fab-tasks', [])):
            if type(fabtask) is dict:
                fabargs = ':'.join(fabtask.items()[0])
                fabargs = parse_arguments([fabargs])[0]
            else:
                fabargs = (fabtask, [], {}, [], [], [])

            name, fargs, kwargs, hosts, roles, excludes = fabargs
            log.info('%02i %s', i, name, extra=dict(module='fab-task'))
            log.debug('args: %s', fabargs, extra=dict(module='fab-task'))
            execute(name, hosts=hosts, roles=roles, exclude_hosts=excludes,
                    *fargs, **kwargs)

    if 'tasks' in data:  # start fabric
        with show(*everything):
            try:
                results = execute(steamroll, data, inspectors, context)
            except KeyboardInterrupt:
                if get_cursor_pos() > 1: print  # prevent log on existing line
                log.warn(u'Killed by Ctrl-C, shutting down…')
                exit_status = 130  # from bash docs

            log.addHandler(locfhandler)
            disconnect_all()

    # exit stage left
    log.note('results %s', results)
    sys.stdout, sys.stderr = orig_stdout, orig_stderr # restore
    logging.shutdown()
    return exit_status


def prep_environ(data, opts):
    'Prepare execution environment according to options given.'
    # look at tasks
    if opts.command:    # override
        data['tasks'] = [opts.command]
    elif not (get(data, '/tasks') or get(data, '/fab-tasks')):
        fatal_exit('No tasks or fab-tasks section found.', os.EX_DATAERR)
    elif opts.select:  # filter tasks, see setup() for err checking
        tasks = data['tasks']
        data['_orig_task_length'] = len(tasks) # save for later
        try:  tasks = safe_eval('tasks[%s]' % opts.select, tasks=tasks)
        except IndexError:
            fatal_exit('-s selection index out of range.', os.EX_USAGE)
        if not tasks:
            fatal_exit('-s no tasks left, holmes.', os.EX_USAGE)
        data['tasks'] = sequence(tasks)

    # explicitly set fabric env keys
    env.warn_only =  True  # handle errors and display to user
    for key, val in (get(data, '/main/env') or {}).iteritems():
        setattr(env, key, val)

    # add to module path
    path = get(data, '/main/sys.path')
    if path:
        if isinstance(path, basestring):
            path = path.split(':')
        sys.path += path

    # set logging folder:
    log_to = get(data, '/main/log-to')
    if log_to:
        global logdir
        logdir = expanduser(log_to)

    # find and expand targeted hosts
    targets = get(data, '/main/targets')
    if isinstance(targets, basestring):
        targets = targets.split()
    if not targets:
        fatal_exit('No targets found.', os.EX_CONFIG)
    targets = expand_targets(targets, data.get('target-groups', []))

    # misc
    env.hosts = targets
    env.pave_bak = get(data, '/main/bak-files')
    env.pave_platform_details = {}
    user = get(data, '/main/user')
    if user and user != CURRENT:
        env.user = user

    # load and minimize inspector script
    inspectors = []
    mode = get(data, '/main/inspect')
    if mode:
        if mode is True:
            filenames = ['inspector.py', 'inspector.sh']
        else:
            filenames = ['inspector.' + mode]
        for filename in filenames:
            script = ''
            with file(join(dirname(__file__), filename)) as f:
                for line in f:  # minimize
                    if line.isspace() or line.startswith('#'):
                        continue
                    script += line.replace('    ', ' ')

            if filename.endswith('.py'):        # add two inspectors to handle
                shell, shellcode = 'python2 -c', script      # lack of python2
                inspectors.append((shell, shellcode))
                shell, shellcode = 'python -c', script
            else:
                shell, shellcode = 'sh -c', script
            inspectors.append((shell, shellcode))

    # set fabric opts
    env.pave_raise_errs = not get(data, '/main/warn-only')
    # not sure if a good idea, but can't see why this would not be set:
    # tell sudo to set the HOME env variable when run as user:
    env.sudo_prefix = "sudo -SHp '%(sudo_prompt)s' " % env
    jobs = get(data, '/main/jobs')
    if jobs > 1:
        log.info('multiple jobs enabled (%s)', jobs)
        env.parallel = True
        env.pool_size = jobs

    # try this... still doesn't separate stderr
    env.combine_stderr = False

    # configure ssh
    keygen()  # generate a local key if there is none
    check_ssh_config()

    return data, inspectors


def setup():
    'Parse command-line, set up logging, create temp folders.'
    global opts
    parser = OptionParser(usage=__doc__.rstrip() % __version__,
                          version=__version__)
    parser.add_option('-b', '--brace-exp', action='store_true',
        help='Use string.format-style brace expansion instead of printf.')
    parser.add_option('-c', '--command', metavar='STR',
        help='Override pavefile tasks with an ad hoc command.')
    parser.add_option('-f', '--filename', metavar='NAME',
        default='.%s%s' % (os.sep, _def_filename),
        help='Use an input file besides %default.')
    parser.add_option('-i', '--interactive', action='store_true',
        help='Stop and ask whether to run each task group.')
    parser.add_option('-j', '--jobs', metavar='#', default=def_jobs,type='int',
        help='Run pave tasks in parallel with the given pool size.')
    parser.add_option('-o', '--option', metavar='N V', action='append',nargs=2,
        help='Override a pavefile option.', default=[])
    parser.add_option('-s', '--select', metavar='#', help='Select task groups'
        ' to run, instead of all. (0-based slice syntax, # or #:#).')
    parser.add_option('-S', '--skel', action='store_true',
        help='Output a "skeleton" pavefile to get started on.')
    parser.add_option('-t', '--target', metavar='T',action='append',default=[],
        help='Override targets with this host or group.  Multiple accepted.')
    parser.add_option('-q', '--quiet', action='store_true',
        help='Silence routine output, allowing warnings and errors.  ' +
             'Overrides verbose.')
    parser.add_option('-v', '--var', metavar=' N V', action='append', nargs=2,
        help='Override a pavefile variable.', default=[])
    parser.add_option('-V', '--verbose', action='store_true',
        help='Enable overwhelmingly verbose debugging output.')

    choices=('auto', 'on', 'off')
    parser.add_option('--color', type='choice', default='auto',
        choices=choices, metavar='CHOICE',
        help='Message coloring/icons %s.' % (choices,))
    parser.add_option('--crypt', action='store_true',
        help='Ask for a password, print its hash and exit. For use with users:'
             ' module in the pavefile.')
    parser.add_option('--speak', action='store_true',
        help='Talk to me, holmes.')
    parser.add_option('--test', action='store_true',
        help='Test mode: parse and validate, skip execution.')

    try:
        opts, args = parser.parse_args()
    except SystemExit:
        sys.exit(os.EX_USAGE)

    # configure console logging
    if opts.quiet or opts.jobs > 1:
        console.setLevel(logging.WARNING)
    else:
        console.setLevel((logging.DEBUG if opts.verbose else logging.INFO))
        if console.level <= logging.DEBUG:
            console.addFilter(SecshFilterD())
    if opts.color == 'auto':
        if console.stream.isatty():     fmtr = ColorFmtr(logfmt)
        else:                           fmtr = logging.Formatter(logfmt)
    elif opts.color in 'on':            fmtr = ColorFmtr(logfmt)
    else:                               fmtr = logging.Formatter(logfmt)
    console.setFormatter(fmtr)

    opts.option =  dict(opts.option)
    if opts.crypt:
        result = ask_crypt()
        print
        if result:
            print >> sys.stderr, '    ', result, '\n'
        sys.exit()

    if opts.skel: # output skeleton file
        import shutil
        dest = join(os.getcwd(), _def_filename)
        if os.path.exists(dest):
            fatal_exit('"%s" already exists.' % dest, os.EX_CANTCREAT)
        try:
            shutil.copyfile(join(dirname(__file__), 'skeleton.yml'), dest)
            log.info('created "%s".', _def_filename)
        except IOError, e:
            fatal_exit('pavefile not created: ' + e.strerror, os.EX_CANTCREAT)
        sys.exit()

    # error checking
    if not os.access(opts.filename, os.R_OK):
        fatal_exit('pavefile "%s" not found or not accessible.'
                    % opts.filename, os.EX_NOINPUT, parser)

    if opts.command and opts.select:
        fatal_exit('-c/--command and -s/--select cannot be used together.',
                    os.EX_USAGE, parser)
    if opts.interactive and opts.jobs > 1:
        fatal_exit('-i/--interactive and -j/--jobs cannot be used together.',
                    os.EX_USAGE, parser)
    if opts.select:
        charset = string.digits + '-' + ':'
        for char in opts.select:
            if char not in charset:
                fatal_exit('-s/--select: invalid character entered.',
                            os.EX_USAGE, parser)
        try:  safe_eval('t[%s]' % opts.select, t=[])
        except IndexError: pass
        except Exception, e:
            fatal_exit('-s, --select: %s' % e, os.EX_USAGE, parser)

    # create temp folders
    folders = (tempdir, 0750), (logdir, 0750)
    for folder, perms in folders:
        try:
            os.makedirs(folder)  # perms incorrect w/makedirs
            if not xdg:
                os.chmod(folder, perms)
        except OSError, e:
            if e.errno == errno.EEXIST: pass
            else:
                fatal_exit("Can't create temp folder: %s" % e, os.EX_CANTCREAT)

    # more logging
    start_file_log(join(logdir, 'localhost'))
    if opts.jobs > 1:   logf = log.note  # promote message
    else:               logf = log.info
    logf('pave version %s, Python: %s, fabric: %s\n' +
         '                  pavefile: %s, logs: %s/logs',
        __version__, platform.python_version(), fabver, opts.filename, tempdir)
    if opts.speak:  # after versions
        log.addHandler(EspeakHandler())

    # check localhost
    with hide('stdout', 'running'):
        platdetails = local(platforms.OS.python %
                        join(dirname(__file__), 'inspector.py'), capture=True)
        if platdetails:
            log.debug(platdetails)
            platdetails = json.loads(platdetails)
            global lcmds
            lcmds = get_platform_cmds(platdetails, local=True)
        else:
            fatal_exit("Inspection of localhost failed." % e, os.EX_SOFTWARE)

    return opts, args


def steamroll(data, inspectors, context):
    '''Let's get rollin...'''
    hostlogh = start_file_log(join(logdir, env.host))
    tasks = data['tasks']
    numtasks = data.get('_orig_task_length') or len(tasks) # not working as expected
    log.note(u'paving %s@%s with %s tasks…', env.user, env.host, len(tasks))

    platdetails = None
    for inspector in inspectors:
        shell, shellcode = inspector
        with hide('running'):
            with settings(warn_only=True, shell=shell):
                try:
                    platdetails = run(shellcode)
                    if platdetails.failed:
                        log.warn(_fmt_log(shell, platdetails.return_code))
                        platdetails = None; continue
                    platdetails = json.loads(platdetails)
                    break
                except NetworkError, e:
                    log.fatal('%s: %s', e.__class__.__name__, e)
                    return 0, 1  # os.EX_NOHOST

    if platdetails is None:
        log.fatal('platform details not found, exiting.')
        return 0, 1

    context = context.copy()  # prevent context spills w/ change below
    context.encoding = platdetails.get('encoding', _def_encoding)
    cmds = get_platform_cmds(platdetails)
    env.pave_platform_details[env.host_string] = platdetails    # make avail
    env.pave_platform = cmds                                    # to mods
    # var expansion for platform details could be done here

    results = []
    offset = 0
    if opts.select:  # fix task numbers
        offset = int(opts.select.partition(':')[0] or 0) # find start

    for i, task in enumerate(tasks):
        i = i + offset
        mycontext = context.copy()  # prevent context spills

        if isinstance(task, basestring):  # handle bare strings as "run" tasks
            mname, sectdata = 'run', task
            desc = mname + ': ' + sectdata[:25]
        elif type(task) is dict:
            task = copy.deepcopy(task)        # prevent mod issues
            mname, sectdata = task.items()[0] # grab the first and only
            desc = mname + ':'
        try:
            log.debug('=' * 60)
            if opts.interactive:
                msg = bold(u'Task group %02i/%02i: %s… [Y/n/a/q]? '
                           % (i, numtasks-1, desc))
                if opts.verbose:
                    print >> sys.stderr, msg,
                    sys.stderr.flush(interactive=True)
                else:
                    log.note(msg)
                res = wait_key(); print >> sys.stderr, res # to console
                if   res == 'q':  break
                elif res == 'n':  continue
                elif res == 'a':  opts.interactive = False
                else:             pass
            results.append(
                get_module(mname).handle(sectdata, cmds, mycontext)
            )
            log.debug('completed task group %s of %s', i, numtasks-1)
        except ImportError, e:
            log.error('could not find module named: %s', e.message.split()[-1])
            log.warn(u'%s shutting down early—error occurred.', env.host)
            results.append( (0,1) )  # add err
            break
        except RuntimeError, e:  # from modules, problem already logged.
            log.warn(u'%s shutting down early—error occurred.', env.host)
            results.append( (0,1) )  # add err
            break
        except Exception, e:
            if log.isEnabledFor(logging.DEBUG):
                log.exc('Uncaught %s:\n  %s' % (e.__class__.__name__,
                                         traceback.format_exc()) )
            else:  # normal means didn't work, perhaps bc of multiproc?
                t = traceback.extract_tb(sys.exc_info()[2])[-1]
                fname = '.' + os.sep + os.sep.join(t[0].split(os.sep)[-3:])
                log.exc( 'Uncaught %s: %s, line %s of %s' % (
                          e.__class__.__name__, e, t[1], fname) )
            log.warn(u'%s shutting down early—exception occurred.', env.host)
            results.append( (0,1) )  # add err
            break

    log.debug('disconnecting.')
    log.removeHandler(hostlogh)
    return sum( c for c,e in results ), sum( e for c,e in results )




import sys, os
from os.path import join
from distutils.core import setup

from pave.meta import __version__, __progname as name

# readme is needed at register/upload time, not install time
try:
    with open('readme.rst') as f:
        long_description = ['|\n', '|\n', '|\n','\n'] # push logo under login box
        for line in f:
            if line.startswith('.. '):          # remove sphinx directives at end
                break
            long_description.append(line)
        long_description = ''.join(long_description)
except IOError:
    long_description = ''

# do we need importlib?
requires = []
if sys.version < '2.7':
    requires.append('importlib')

# install helper script?
scripts = [join(name, name)]
if os.name == 'nt':
    scripts.append(name + '.cmd')


setup(
    name              = name,
    version           = __version__,
    description       = ( 'Simple push-based configuration and deployment ' +
                          'tool, leveraging fabric.  ' +
                          'No servers, few dependencies.'),
    author            = 'Mike Miller',
    author_email      = 'mixmastamyk@bitbucket.org',
    url               = 'https://bitbucket.org/mixmastamyk/pave',
    download_url      = 'https://bitbucket.org/mixmastamyk/pave/get/' +
                        'default.tar.gz',
    license           = 'GPLv3+',
    requires          = ['PyYAML(>=3.05,<4.0)', 'fabric(>=1.4,<2.0)',
                         'voluptuous(>=0.8.3,<0.9)'] + requires,# for pypi page
    install_requires  = ['PyYAML>=3.05,<4.0a0', 'fabric>=1.4,<2.0a0',
                         'voluptuous>=0.8.3,<0.9a0'] + requires,  # real reqs
                         #~ 'voluptuous==dev'],
    packages          = [name, join(name, 'lib')],
    scripts           = scripts,
    package_data      = {name: ['inspector.sh', 'skeleton.yml']},
    #~ dependency_links  = ['https://github.com/alecthomas/voluptuous/tarball/' +
                         #~ 'master#egg=voluptuous-dev'],  # temp dev download
    extras_require = {
        'win': ['colorama', 'fcrypt'],
    },

    long_description  = long_description,
    classifiers       = [
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: System :: Installation/Setup',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
    ],
)



# (C) 2012-2013 Mike Miller.
# License: GNU GPLv3+
'''
    Manage system daemons/services.  Values are whitespace separated strings.

    ``restart:``, ``restart-fail:``
        restart will start the service if it was stopped, restart-fail will
        fail with an error if it was not running originally.

    .. note::

        Some daemon startup scripts are not written well and quit when the
        ssh session ends.  To prevent this from occurring, prepend the
        service name with "``nohup:``" as shown below.

    Example::

        - services:
            # the follwing two are temporary until reboot:
            running: hostname hwclock
            stopped: whoopsie
            # permanently; /packages/remove more reliable on ubuntu
            disabled: whoopsie
            restart: nohup:memcached

'''
import logging

from fabric.api import env, run
from fabric.contrib.files import exists as rexists
from voluptuous import Required

from pave.schema import conditionals_schema
from pave.utils import runcmd, eval_conditionals, err_cond


log = logging.getLogger()
val = basestring
req = lambda key: Required(key, default='')
schema = {
    req('running'): val, req('stopped'): val,       req('disabled'): val,
    req('restart'): val, req('restart-fail'): val,
}
schema.update(conditionals_schema)


def check_nohup(svcname):
    'look for a nohup:SVCNAME construct.'
    if ':' in svcname:
        nohup, _, svcname = svcname.partition(':')
        nohup += ' '
    else:
        nohup = ''
    return nohup, svcname


def handle(section, cmds, context):
    changed, errors = 0, 0
    if eval_conditionals(section, context) is False:
        log.skip(err_cond, '')  # log here to tag from this module
        return changed, errors

    # query current status
    services, running = {}, []
    svclist = runcmd(cmds.svc_lst)  # doesn't need to run as root
    if svclist.failed:
        log.error(svclist); errors += 1 # don't quit on test failure
    else:
        services = cmds.parse_svclist(svclist) # a dict
        sorted_services = sorted(services.items())
        if log.isEnabledFor(logging.DEBUG):
            svctxt = ', '.join([ ('%s:%s' % svc) for svc in sorted_services])
            log.debug('existing: %s, length:%s',  '{' + svctxt + '}',
                      len(services))
        running = [ s  for s,st in sorted_services  if st == 1 ]

    # loop over section, instead of below:
    shouldruns = section.get('running', '').split()
    for svcname in shouldruns:
        nohup, svcname = check_nohup(svcname)
        if svcname not in running:
            log.change('starting %s', svcname)
            result = runcmd(nohup + cmds.svc_start, svcname, **context)
            if result.succeeded:   changed += 1
            else:
                log.error(result); errors += 1
                if env.pave_raise_errs: raise RuntimeError
        elif svcname not in services:
            log.error('%s not installed.', svcname); errors += 1
        else:
            log.skip('%s already started', svcname)

    shouldstops = section.get('stopped', '').split()
    for svcname in shouldstops:
        if svcname in running:
            log.change('stopping %s', svcname)
            result = runcmd(cmds.svc_stop, svcname, **context)
            if result.succeeded:   changed += 1
            else:
                log.error(result); errors += 1
                if env.pave_raise_errs: raise RuntimeError
        elif svcname not in services:
            log.error('%s not installed.', svcname); errors += 1
        else:
            log.skip('%s already stopped.', svcname)

    shoulddiss = section.get('disabled', '').split()
    for svcname in shoulddiss:  # not entirely sure how this should work
        found = False
        if run('ls /etc/rc?.d/S*%s*' % svcname).succeeded:
            found = True
            log.change('disabling rc.d %s', svcname)
            result = runcmd(cmds.svc_disable, svcname, **context)
            if result.succeeded:   changed += 1
            else:
                log.error(result); errors += 1
                if env.pave_raise_errs: raise RuntimeError
        elif run('ls /etc/rc?.d/K*%s*' % svcname).succeeded:
            found = True
            log.skip('rc.d %s already disabled.', svcname)

        if hasattr(cmds, 'svc_disable_ups'):
            if (rexists('/etc/init/%s.disabled' % svcname) and not
                rexists('/etc/init/%s.conf' % svcname)):
                found = True
                log.skip('upstart job %s already disabled.', svcname)
            elif rexists('/etc/init/%s.conf' % svcname):
                found = True
                log.change('disabling upstart job %s', svcname)
                result = runcmd(cmds.svc_disable_ups, svcname, svcname,
                                svcname, **context)
                if result.succeeded:   changed += 1
                else:
                    log.error(result); errors += 1
                    if env.pave_raise_errs: raise RuntimeError

        if not found:
            log.error('%s not installed.', svcname); errors += 1
            if env.pave_raise_errs: raise RuntimeError

    shouldrests = section.get('restart-fail', '').split()
    for svcname in shouldrests:
        nohup, svcname = check_nohup(svcname)
        if svcname in running:
            log.change('restarting %s', svcname)
            result = runcmd(nohup + cmds.svc_restart, svcname, **context)
            if result.succeeded:   changed += 1
            else:
                log.error(result); errors += 1
                if env.pave_raise_errs: raise RuntimeError
        elif svcname not in services:
            log.error('%s not installed.', svcname); errors += 1
        else:
            log.error('%s not running, can\'t be restarted.', svcname);
            errors += 1
            if env.pave_raise_errs: raise RuntimeError

    shouldrests = section.get('restart', '').split()
    for svcname in shouldrests:
        nohup, svcname = check_nohup(svcname)
        if svcname in running:
            log.change('restarting %s', svcname)
            result = runcmd(nohup + cmds.svc_restart, svcname, **context)
            if result.succeeded:   changed += 1
            else:
                log.error(result); errors += 1
                if env.pave_raise_errs: raise RuntimeError
        elif svcname not in services:
            log.error('%s not installed.', svcname); errors += 1
        else:
            log.change('restart: starting stopped %s', svcname)
            result = runcmd(nohup + cmds.svc_start, svcname, **context)
            if result.succeeded:   changed += 1
            else:
                log.error(result); errors += 1
                if env.pave_raise_errs: raise RuntimeError

    return changed, errors




# (c) 2012, Zettar Inc.
# Written by Chin Fang <fangchin@zettar.com>
#   https://github.com/ansible/ansible/tree/devel/lib/ansible/inventory
# regex implementation and doctests by Mike Miller
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.
#
'''
    This module is for enhancing the inventory parsing capability such
    that it can deal with hostnames specified using a simple pattern in the
    form of [beg:end], example: [1:5] where if begin is not specified, it
    defaults to 0.

    If beg is given and is left-zero-padded, e.g. '001', it is taken as a
    formatting hint when the range is expanded. e.g. [001:010] is to be
    expanded into 001, 002 ...009, 010.

    Note that when beg is specified with left zero padding, then the length of
    end must be the same as that of beg, else a exception is raised.
'''
import re

class HostSpecError(Exception):
    pass

range_fmt = re.compile(r'''
    (                       # group host basename
        [A-Za-z0-9\-\.]*    # zero or more alphanumeric, dash, or period
    )

    \[              # A left bracket
    (               # group start digits
        \d*         # zero or more digits
    )
        [:\-]       # a colon or dash
    (               # group end digits
        \d+         # one or more digits
    )
    \]              # A right bracket
    (               # group host domain
        [\w\-\.]*   # zero or more alphanumeric, underscore, dash, or period
    )
''', re.VERBOSE) # |re.DEBUG) prints regex details


def detect_range(hostspec):
    '''
        A helper function that checks a given host string to see if it contains
        a range pattern descibed in the docstring above.

        Returns a True match object if the given string contains a pattern,
        else None.

        >>> detect_range('hostname4')

        >>> detect_range('192.[168].2.1')

        >>> detect_range('foo[bar]baz')

        >>> detect_range('[@$\\n*#:*\\t$&@]\\]')

        >>> detect_range('host_name[1:8]')  # no underscores in hostnames

        >>> detect_range('host[1:8]').groups()
        ('host', '1', '8', '')

        >>> detect_range('10.0.0.1[01:08]').groups()
        ('10.0.0.1', '01', '08', '')

        >>> detect_range('host[1:8].with-dash.com').groups()
        ('host', '1', '8', '.with-dash.com')

        >>> detect_range('www-[0001:8000].g.cn').groups()
        ('www-', '0001', '8000', '.g.cn')
    '''
    return range_fmt.match(hostspec)


def expand_hostname_range(matchobj):
    '''
        A helper function that expands a given string that contains a pattern
        specified in top docstring, and returns a list that consists of the
        expanded version.

        The '[' and ']' characters are used to maintain the pseudo-code
        appearance. They are replaced in this function with '|' to ease
        string splitting.

        Example::

            Passing a match obj allows a second execution of the regex to be
            skipped:
            match = detect_range(hostspec)
            if match:
                expand_hostname_range(match)

        References: http://ansible.github.com/patterns.html#hosts-and-groups

        >>> expand_hostname_range(detect_range('host[1:8]'))
        ['host1', 'host2', 'host3', 'host4', 'host5', 'host6', 'host7', 'host8']

        >>> expand_hostname_range(detect_range('host[0-2]'))
        ['host0', 'host1', 'host2']

        >>> expand_hostname_range(detect_range('host[0:3]'))
        ['host0', 'host1', 'host2', 'host3']

        >>> expand_hostname_range(detect_range('host[:2]'))
        ['host0', 'host1', 'host2']

        >>> expand_hostname_range(detect_range('www[01:04].g.cn'))
        ['www01.g.cn', 'www02.g.cn', 'www03.g.cn', 'www04.g.cn']

        >>> expand_hostname_range(detect_range('[0001:0002]node.g.cn'))
        ['0001node.g.cn', '0002node.g.cn']
    '''
    all_hosts = []
    head, beg, end, tail = matchobj.groups()
    if not beg:
        beg = '0'
    if not end:
        raise HostSpecError('Host range end value missing.')

    # range length formatting hint
    rlen = 0
    if beg[0] == '0' and len(beg) > 1:
        rlen = len(beg)

    if rlen and rlen != len(end):
        raise HostSpecError('Host range format incorrectly specified.')

    for i in range(int(beg), int(end)+1):
        if rlen:
            rseq = str(i).zfill(rlen) # range sequence
        else:
            rseq = str(i)
        all_hosts.append(head + rseq + tail)

    return all_hosts


if __name__ == '__main__':
    import doctest
    doctest.testmod()




# (C) 2012-2013 Mike Miller.
# License: GNU GPLv3+
'''
    Copy files to and fro.  Given a source and destination, copy files.

    Notes:

    * By default a put (upload) is performed, but the direction can be
      specified as the first parameter.
    * Paths with whitespace must be quoted.
    * Destinations folders should end in a ``/`` character for efficiency,
      as doing so skips the need for several filesystem tests.
    * If sudo is enabled,
      relative paths/tildes will be interpreted (and files owned) by the
      root user.
      Therefore it is recommended that absolute remote paths be given.
    * Does not currently handle wildcards.
    * While this module is similar to using the scp command-line program it
      may not work identically.

    Example::

        - scp:
            - SRC DEST          # put
            - SRC DEST MODE     # put, w/ Unix octal mode or "same"
            - get SRC DEST

            # example conditional
            - if: inst_ssl == True
              do: put bundle.crt "%sitefldr/ssl/bundle.crt"

    Additional details:

    * `fabric.operations.get <http://docs.fabfile.org/en/latest/api/core/operations.html#fabric.operations.get>`_
    * `fabric.operations.put <http://docs.fabfile.org/en/latest/api/core/operations.html#fabric.operations.put>`_

'''
import logging
from os.path import basename, join

from fabric.api import env
from fabric.operations import get, local, put
from voluptuous import Invalid, All

from pave.schema import conditionals_schema
from pave.utils import q, runcmd, shplit, check_list_conditionals

log = logging.getLogger()


def check(v):
    'validate command lines.  Todo: should check if on disk.'
    tokens = shplit(v)
    length = len(tokens)
    if tokens[0] == 'get':
        if length != 3:
            raise Invalid('Expected length of 3.')
    else:
        if (length < 2) or (length > 4):
            raise  Invalid('Expected length of 2-4.')
        if length > 3 and '=' not in tokens[3]:
            raise Invalid('Expected mode=XXXX format.')
    return v
#~ subtask_schema = {basestring: All(basestring, check)}
subtask_schema = {basestring: basestring}
subtask_schema.update(conditionals_schema)

schema = [  All(basestring, check), subtask_schema ]


def tilde_helper(path1, path2):
    if path1.startswith('~'):
        path1 = '$HOME' + path1[1:]
    if path2.startswith('~'):
        path2 = '$HOME' + path2[1:]
    return path1, path2


def handle(section, cmds, context):
    changed, errors = 0, 0
    context.pop('sudo_user', None)
    context.pop('encoding', None)
    def hash_files(locfn, remfn):
        errors = 0
        result = runcmd(cmds.file_hash, remfn, **context)
        if result.succeeded:
            remhash = result.split()[0]
        else:
            log.error(result); errors += 1

        result = local(cmds.file_hash % locfn, capture=True)
        if result.succeeded:
            lochash = result.split()[0]
            log.debug('local hash: %s', lochash)
        else:
            log.error(result); errors += 1
        return lochash, remhash, errors


    for command in section:
        command = check_list_conditionals(command, context, module=__name__)
        if not command: continue

        args = shplit(command)
        if args[0] == 'put':
            mode = 'put'
            args = args[1:]
        elif args[0] == 'get':
            mode = 'get'
            args = args[1:]
        else:
            mode = 'put'

        if mode == 'put':
            mode, mycon = None, {}  # overloading mode, not a great idea
            if len(args) == 2:
                locfn, remfn = args
            elif len(args) == 3:
                locfn, remfn, mode = args
                mode = mode.partition('=')[2]
            locfn, remfn = tilde_helper(locfn, remfn)

            # check if file mode should be set
            if mode == 'same':
                mycon = context.copy()
                mycon.mirror_local_mode = True
                mode = None # reset to put default
            elif mode and mode.isdigit():
                mode = int(mode, 8)
                mycon = context
            else:
                mycon = context

            # if the remfn ends in /, add filename.
            lbase = basename(locfn)
            rbase = basename(remfn)
            if lbase and not rbase:
                remfn = join(remfn, lbase)

            # check for existing and identical file
            # don't halt on test failure
            if runcmd(cmds.file_exists, remfn, **context).succeeded:
                log.debug('"%s" found.', remfn)

                lochash, remhash = None, None
                if rbase:                  # is this a folder?
                    if runcmd(cmds.dir_test, remfn, **context).succeeded:
                        log.debug('"%s" is a folder.', remfn)
                        remfn = join(remfn, basename(locfn))

                        if runcmd(cmds.file_exists,remfn, **context).succeeded:
                            log.debug('"%s" found.', remfn)
                            lochash, remhash, e = hash_files(locfn, remfn)
                        else:
                            log.debug('remote file not found. %s', remfn)
                    else:
                        lochash, remhash, e = hash_files(locfn, remfn)
                else:
                    lochash, remhash, e = hash_files(locfn, remfn)

                if lochash and (lochash == remhash):
                    log.skip('put: not uploading identical file: %s.',q(locfn))
                    continue
            else:
                log.debug('remote file not found. %s', remfn)

            # upload the file
            log.change('put: %s --> %s', q(locfn), q(remfn))
            try:
                result = put(locfn, remfn, mode=mode, **mycon)  # returns list
            except ValueError, e:  # :/
                log.error(str(e)); errors += 1
                if env.pave_raise_errs: raise RuntimeError
            if result.succeeded:  changed += 1
            else:
                log.error('put failed--check permissions.'); errors += 1
                if env.pave_raise_errs: raise RuntimeError

        elif mode == 'get':
            remfn, locfn = args
            locfn, remfn = tilde_helper(locfn, remfn)
            # copy filename to remote if nec.
            lbase = basename(locfn)
            rbase = basename(remfn)
            if rbase and not lbase:
                locfn = join(locfn, rbase)

            # check for existing and identical file, dirs not yet supported
            # don't halt on test failure
            if local(cmds.file_test % locfn, capture=True).succeeded:
                remhash, lochash = None, None
                result = runcmd(cmds.file_hash, remfn, **context)
                if result.succeeded:
                    remhash = result.split()[0]
                else:
                    log.error(result); errors += 1

                result = local(cmds.file_hash % locfn, capture=True)
                if result.succeeded:
                    lochash = result.split()[0]
                else:
                    log.error(result); errors += 1

                if lochash and (lochash == remhash):
                    log.skip('get: not downloading identical file: ' +
                             '%s.', q(remfn))
                    continue
            else:
                log.debug('local file not found. %s', locfn)

            # download the file
            log.change('get: %s <--- %s', q(locfn), q(remfn))
            result = get(remfn, locfn)  # returns a list
            if result.succeeded:  changed += 1
            else:
                log.error('get failed--check permissions.'); errors += 1
                if env.pave_raise_errs: raise RuntimeError

    return changed, errors

